# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the mask module.
"""

import astropy.units as u
import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal
import pytest

from ..bounding_box import BoundingBox
from ..mask import RegionMask
from ..pixcoord import PixCoord
from ...shapes import CirclePixelRegion, CircleAnnulusPixelRegion

POSITIONS = [(-20, -20), (-20, 20), (20, -20), (60, 60)]


def test_mask_input_shapes():
    with pytest.raises(ValueError):
        mask_data = np.ones((10, 10))
        bbox = BoundingBox(5, 10, 5, 10)
        RegionMask(mask_data, bbox)


def test_mask_array():
    mask_data = np.ones((10, 10))
    bbox = BoundingBox(5, 15, 5, 15)
    mask = RegionMask(mask_data, bbox)
    data = np.array(mask)
    assert_allclose(data, mask.data)


def test_mask_get_overlap_slices():
    aper = CirclePixelRegion(PixCoord(5, 5), radius=10.)
    mask = aper.to_mask()
    slc = ((slice(0, 16, None), slice(0, 16, None)),
           (slice(5, 21, None), slice(5, 21, None)))
    assert mask.get_overlap_slices((25, 25)) == slc


def test_mask_cutout_shape():
    mask_data = np.ones((10, 10))
    bbox = BoundingBox(5, 15, 5, 15)
    mask = RegionMask(mask_data, bbox)

    with pytest.raises(ValueError):
        mask.cutout(np.arange(10))

    with pytest.raises(ValueError):
        mask.to_image((10,))


def test_mask_cutout_copy():
    data = np.ones((50, 50))
    aper = CirclePixelRegion(PixCoord(25, 25), radius=10.)
    mask = aper.to_mask()
    cutout = mask.cutout(data, copy=True)
    data[25, 25] = 100.
    assert cutout[10, 10] == 1.

    # test quantity data
    data2 = np.ones((50, 50)) * u.adu
    cutout2 = mask.cutout(data2, copy=True)
    assert cutout2.unit == data2.unit
    data2[25, 25] = 100. * u.adu
    assert cutout2[10, 10].value == 1.


@pytest.mark.parametrize('position', POSITIONS)
def test_mask_cutout_no_overlap(position):
    data = np.ones((50, 50))
    aper = CirclePixelRegion(PixCoord(position[0], position[1]), radius=10.)
    mask = aper.to_mask()

    cutout = mask.cutout(data)
    assert cutout is None

    weighted_data = mask.multiply(data)
    assert weighted_data is None

    image = mask.to_image(data.shape)
    assert image is None


@pytest.mark.parametrize('position', POSITIONS)
def test_mask_cutout_partial_overlap(position):
    data = np.ones((50, 50))
    aper = CirclePixelRegion(PixCoord(position[0], position[1]), radius=30.)
    mask = aper.to_mask()

    cutout = mask.cutout(data)
    assert cutout.shape == mask.shape

    weighted_data = mask.multiply(data)
    assert weighted_data.shape == mask.shape

    image = mask.to_image(data.shape)
    assert image.shape == data.shape


def test_mask_multiply():
    radius = 10.
    data = np.ones((50, 50))
    region = CirclePixelRegion(PixCoord(25, 25), radius=radius)
    mask = region.to_mask(mode='exact')
    data_weighted = mask.multiply(data)
    assert_almost_equal(np.sum(data_weighted), np.pi * radius**2)

    # test that multiply() returns a copy
    data[25, 25] = 100.
    assert data_weighted[10, 10] == 1.


def test_mask_multiply_quantity():
    radius = 10.
    data = np.ones((50, 50)) * u.adu
    region = CirclePixelRegion(PixCoord(25, 25), radius=radius)
    mask = region.to_mask(mode='exact')
    data_weighted = mask.multiply(data)
    assert data_weighted.unit == u.adu
    assert_almost_equal(np.sum(data_weighted.value), np.pi * radius**2)

    # test that multiply() returns a copy
    data[25, 25] = 100. * u.adu
    assert data_weighted[10, 10].value == 1.


@pytest.mark.parametrize('value', (np.nan, np.inf))
def test_mask_nonfinite_fill_value(value):
    region = CircleAnnulusPixelRegion(PixCoord(0, 0), 10, 20)
    data = np.ones((101, 101)).astype(int)
    cutout = region.to_mask().cutout(data, fill_value=value)
    assert ~np.isfinite(cutout[0, 0])


def test_mask_multiply_fill_value():
    region = CircleAnnulusPixelRegion(PixCoord(0, 0), 10, 20)
    data = np.ones((101, 101)).astype(int)
    cutout = region.to_mask().multiply(data, fill_value=np.nan)
    xypos = ((20, 20), (5, 5), (5, 35), (35, 5), (35, 35))
    for x, y in xypos:
        assert np.isnan(cutout[y, x])


def test_mask_nonfinite_in_bbox():
    """
    Regression test that non-finite data values outside of the mask but
    within the bounding box are set to zero.
    """
    data = np.ones((101, 101))
    data[33, 33] = np.nan
    data[67, 67] = np.inf
    data[33, 67] = -np.inf
    data[22, 22] = np.nan
    data[22, 23] = np.inf

    radius = 20.
    reg1 = CirclePixelRegion(PixCoord(50, 50), radius)
    reg2 = CirclePixelRegion(PixCoord(5, 5), radius)

    wdata1 = reg1.to_mask(mode='exact').multiply(data)
    assert_allclose(np.sum(wdata1), np.pi * radius**2)

    wdata2 = reg2.to_mask(mode='exact').multiply(data)
    assert_allclose(np.sum(wdata2), 561.6040111923013)


@pytest.mark.parametrize('x, y, exp_shape',
                         [(0, 0, 245), (50, 50, 940), (100, 100, 245)])
def test_mask_get_values(x, y, exp_shape):
    aper = CircleAnnulusPixelRegion(PixCoord(x, y), inner_radius=10,
                                    outer_radius=20)
    data = np.ones((101, 101))
    values = aper.to_mask(mode='center').get_values(data)
    assert values.shape == (exp_shape,)
    assert_allclose(np.sum(values), exp_shape)


def test_mask_get_values_no_overlap():
    aper = CirclePixelRegion(PixCoord(-100, -100), radius=3)
    data = np.ones((51, 51))
    values = aper.to_mask().get_values(data)
    assert values.shape == (0,)


def test_mask_get_values_mask():
    aper = CirclePixelRegion(PixCoord(24.5, 24.5), radius=10.)
    data = np.ones((51, 51))
    mask = aper.to_mask(mode='exact')
    with pytest.raises(ValueError):
        mask.get_values(data, mask=np.ones(3))

    arr = mask.get_values(data, mask=None)
    assert_allclose(np.sum(arr), 100. * np.pi)

    data_mask = np.zeros(data.shape, dtype=bool)
    data_mask[25:] = True
    arr2 = mask.get_values(data, mask=data_mask)
    assert_allclose(np.sum(arr2), 100. * np.pi / 2.)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
The module provides several custom descriptor classes for attribute
validation of region classes.
"""

import abc

from astropy.coordinates import SkyCoord
from astropy.units import Quantity
import numpy as np

from .pixcoord import PixCoord

__all__ = []


class RegionAttr(abc.ABC):
    """
    Base descriptor class for region attribute validation.
    """

    def __init__(self, name):
        self.name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return instance.__dict__[self.name]

    def __set__(self, instance, value):
        self._validate(value)
        instance.__dict__[self.name] = value

    def __delete__(self, instance):
        del instance.__dict__[self.name]

    @abc.abstractmethod
    def _validate(self, value):
        """
        Validate the attribute value.

        An exception is raised if the value is invalid.
        """
        pass


class ScalarPix(RegionAttr):
    """
    Descriptor class for `~regions.PixelRegion`, which takes a scalar
    `~regions.PixCoord` object.
    """

    def _validate(self, value):
        if not (isinstance(value, PixCoord) and value.isscalar):
            raise ValueError(f'The {self.name} must be a scalar PixCoord '
                             'object')


class OneDPix(RegionAttr):
    """
    Descriptor class for `~regions.PixelRegion`, which takes a
    one-dimensional `regions.PixCoord` object.
    """

    def _validate(self, value):
        if not (isinstance(value, PixCoord) and not value.isscalar
                and value.x.ndim == 1):
            raise ValueError(f'The {self.name} must be a 1D PixCoord object')


class ScalarLength(RegionAttr):
    """
    Descriptor class for `~regions.PixelRegion`, which takes a scalar
    python/numpy number.
    """

    def _validate(self, value):
        if not np.isscalar(value):
            raise ValueError(
                f'The {self.name} must be a scalar numpy/python number')


class ScalarSky(RegionAttr):
    """
    Descriptor class for `~regions.SkyRegion`, which takes a scalar
    `~astropy.coordinates.SkyCoord` object.
    """

    def _validate(self, value):
        if not (isinstance(value, SkyCoord) and value.isscalar):
            raise ValueError(f'The {self.name} must be a scalar SkyCoord '
                             'object')


class OneDSky(RegionAttr):
    """
    Descriptor class for `~regions.SkyRegion`, which takes a
    one-dimensional `~astropy.coordinates.SkyCoord` object.
    """

    def _validate(self, value):
        if not (isinstance(value, SkyCoord) and value.ndim == 1):
            raise ValueError(f'The {self.name} must be a 1D SkyCoord object')


class QuantityLength(RegionAttr):
    """
    Descriptor class for `~regions.SkyRegion`, which takes a scalar
    `~astropy.units.Quantity` object.
    """

    def _validate(self, value):
        if not (isinstance(value, Quantity) and value.isscalar):
            raise ValueError(f'The {self.name} must be a scalar astropy '
                             'Quantity object')


class RegionType(RegionAttr):
    """
    Descriptor class for compound pixel and sky regions.
    """

    def __init__(self, name, regionclass):
        self.name = name
        self.regionclass = regionclass

    def _validate(self, value):
        if not isinstance(value, self.regionclass):
            raise ValueError(f'The {self.name} must be a '
                             f'{self.regionclass.__name__} object')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Geometry subpackage for low-level geometry functions.
"""

from .circular_overlap import *  # noqa
from .elliptical_overlap import *  # noqa
from .rectangular_overlap import *  # noqa
from .polygonal_overlap import *  # noqa



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from astropy.tests.helper import assert_quantity_allclose
from astropy.units.quantity import Quantity
import pytest

from ..core import _to_shape_list
from ..crtf.read import _CRTFParser
from ..ds9.read import _DS9Parser


def test_shape_ds9():
    reg_str = 'galactic\ncircle(42,43,3) # color=green'

    parser = _DS9Parser(reg_str)
    shape1 = parser.shapes[0]
    region = parser.shapes.to_regions()
    shape2 = _to_shape_list(region, 'galactic')[0]

    assert shape1.coordsys == shape2.coordsys
    assert shape1.region_type == shape2.region_type
    assert shape1.include == shape2.include
    assert shape1.coord == shape2.coord
    assert dict(shape1.meta) == dict(shape2.meta)
    assert shape1.composite == shape2.composite

    # Checks for pixel offset since DS9 has a origin=1 pixel system
    reg_str = 'image\nbox(1, 2, 3, 4, 5)'

    parser = _DS9Parser(reg_str)
    shape = parser.shapes[0]

    assert shape.coord[0] == 0
    assert shape.coord[1] == 1
    assert_quantity_allclose(shape.coord[2:-1], [Quantity('3'), Quantity('4')])
    assert_quantity_allclose(shape.coord[-1], Quantity('5deg'))


def test_shape_crtf():
    reg_str = ('circle[[18h12m24s, -23d11m00s], 2.3arcsec], linewidth=2, '
               'coord=J2000, symsize=2')

    parser = _CRTFParser(reg_str)
    shape1 = parser.shapes[0]
    region = parser.shapes.to_regions()
    shape2 = _to_shape_list(region, 'fk5')[0]

    assert shape1.coordsys == shape2.coordsys
    assert shape1.region_type == shape2.region_type
    assert shape1.include == shape2.include
    assert_quantity_allclose(shape1.coord, shape2.coord)
    # assert dict(shape1.meta) == dict(shape2.meta)
    assert shape1.composite == shape2.composite

    # Pixel origin is assumed to be 0 origin. Needs to be checked though.
    reg_str = 'circle[[1pix, 2pix], 3pix], linewidth=2, coord=image, symsize=2'

    parser = _CRTFParser(reg_str)
    shape = parser.shapes[0]
    assert_quantity_allclose(shape.coord,
                             [Quantity('1'), Quantity('2'), Quantity('3')])


def test_valid_shape():
    reg_str = ('circle[[18h12m24s, -23d11m00s], 2.3arcsec], linewidth=2, '
               'coord=J2000, symsize=2')

    shape = _CRTFParser(reg_str).shapes[0]

    with pytest.raises(ValueError) as excinfo:
        shape.region_type = 'box'

    assert '"box" is not a valid region type' in str(excinfo.value)

    shape = _CRTFParser(reg_str).shapes[0]
    with pytest.raises(ValueError) as excinfo:
        shape.coordsys = 'hello'

    estr = '"hello" is not a valid coordinate reference frame'
    assert estr in str(excinfo.value)


def test_valid_ellipse_ds9():
    reg_str1 = 'image\nellipse (1, 2, 3, 4, 5)'
    reg_str2 = 'fk5\nellipse (1, 2, 3, 4, 5)'
    reg_str3 = 'image\nellipse (1, 2, 3, 4, 5, 6, 7)'  # Elliptical Annulus

    shape1 = _DS9Parser(reg_str1, 'warn').shapes[0]
    shape2 = _DS9Parser(reg_str2, 'warn').shapes[0]
    shape3 = _DS9Parser(reg_str3, 'warn').shapes[0]

    assert_quantity_allclose(shape1.coord[:2], [0, 1])
    assert_quantity_allclose(shape2.coord[:2],
                             [Quantity('1deg'), Quantity('2deg')])
    assert_quantity_allclose(shape3.coord[:2], [0, 1])

    assert_quantity_allclose(shape1.coord[2:-1], [6, 8])
    assert_quantity_allclose(shape2.coord[2:-1],
                             [Quantity('6deg'), Quantity('8deg')])
    assert_quantity_allclose(shape3.coord[2:-1], [6, 8, 10, 12])

    assert_quantity_allclose(shape1.coord[-1], Quantity('5deg'))
    assert_quantity_allclose(shape2.coord[-1], Quantity('5deg'))
    assert_quantity_allclose(shape3.coord[-1], Quantity('7deg'))


def test_valid_ellipse_crtf():
    reg_str1 = 'ellipse[[1pix, 2pix], [3pix, 4pix], 5deg]'
    reg_str2 = 'ellipse[[1deg, 2deg], [3deg, 4deg], 5deg], coord=fk5'

    shape1 = _CRTFParser(reg_str1, 'warn').shapes[0]
    shape2 = _CRTFParser(reg_str2, 'warn').shapes[0]

    assert_quantity_allclose(shape1.coord[:2], [1, 2])
    assert_quantity_allclose(shape2.coord[:2],
                             [Quantity('1deg'), Quantity('2deg')])

    assert_quantity_allclose(shape1.coord[2:-1], [8, 6])
    assert_quantity_allclose(shape2.coord[2:-1],
                             [Quantity('8deg'), Quantity('6deg')])

    assert_quantity_allclose(shape1.coord[-1], Quantity('5deg'))
    assert_quantity_allclose(shape2.coord[-1], Quantity('5deg'))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from astropy.utils.exceptions import AstropyUserWarning

__all__ = ['DS9RegionParserWarning', 'DS9RegionParserError']


class DS9RegionParserWarning(AstropyUserWarning):
    """
    A generic warning class for DS9 region parsing.
    """


class DS9RegionParserError(ValueError):
    """
    A generic error class for DS9 region parsing.
    """


# mapping to matplotlib marker symbols, also compatible with CRTF.
valid_symbols_ds9 = {'circle': 'o',
                     'box': 's',
                     'diamond': 'D',
                     'x': 'x',
                     'cross': '+',
                     'arrow': '^',
                     'boxcircle': '*'}



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numbers
import string
from warnings import warn

from astropy.coordinates import (Angle, SkyCoord, UnitSphericalRepresentation,
                                 frame_transform_graph)
from astropy.table import Table
import astropy.units as u
from astropy.utils.decorators import deprecated
import numpy as np

from ..shapes import (CirclePixelRegion, CircleSkyRegion,
                      EllipsePixelRegion, EllipseSkyRegion,
                      RectanglePixelRegion, RectangleSkyRegion,
                      PolygonPixelRegion, PolygonSkyRegion,
                      CircleAnnulusPixelRegion, CircleAnnulusSkyRegion,
                      EllipseAnnulusPixelRegion, EllipseAnnulusSkyRegion,
                      RectangleAnnulusPixelRegion, RectangleAnnulusSkyRegion,
                      LinePixelRegion, LineSkyRegion,
                      PointPixelRegion, PointSkyRegion,
                      TextPixelRegion, TextSkyRegion)
from ..core.core import PixCoord, SkyRegion
from ..core.metadata import RegionMeta, RegionVisual
from .ds9.core import DS9RegionParserWarning, valid_symbols_ds9
from .crtf.core import CRTFRegionParserWarning

__all__ = ['ShapeList', 'Shape']

regions_attributes = dict(circle=['center', 'radius'],
                          ellipse=['center', 'width', 'height', 'angle'],
                          rectangle=['center', 'width', 'height', 'angle'],
                          polygon=['vertices'],
                          circleannulus=['center', 'inner_radius',
                                         'outer_radius'],
                          ellipseannulus=['center', 'inner_width',
                                          'inner_height', 'outer_width',
                                          'outer_height', 'angle'],
                          line=['start', 'end'],
                          point=['center'],
                          text=['center'])

regions_attributes['rectangleannulus'] = regions_attributes['ellipseannulus']

# Map the region names in the respective format to the ones available in
# this package
reg_mapping = {'DS9': {x: x for x in regions_attributes},
               'CRTF': {x: x for x in regions_attributes},
               'FITS_REGION': {x: x for x in regions_attributes}}

reg_mapping['DS9']['box'] = 'rectangle'
reg_mapping['CRTF']['rotbox'] = 'rectangle'
reg_mapping['CRTF']['box'] = 'rectangle'
reg_mapping['CRTF']['centerbox'] = 'rectangle'
reg_mapping['CRTF']['poly'] = 'polygon'
reg_mapping['CRTF']['symbol'] = 'point'
reg_mapping['CRTF']['text'] = 'text'
reg_mapping['CRTF']['annulus'] = 'circleannulus'
reg_mapping['DS9']['text'] = 'text'
reg_mapping['DS9']['annulus'] = 'circleannulus'
reg_mapping['FITS_REGION']['annulus'] = 'circleannulus'
reg_mapping['FITS_REGION']['box'] = 'rectangle'
reg_mapping['FITS_REGION']['rotbox'] = 'rectangle'
reg_mapping['FITS_REGION']['elliptannulus'] = 'ellipseannulus'

# valid astropy coordinate frames in their respective formats
valid_coordsys = {'DS9': ['image', 'physical', 'fk4', 'fk5', 'icrs',
                          'galactic', 'geocentrictrueecliptic', 'wcs'],
                  'CRTF': ['image', 'fk5', 'fk4', 'galactic',
                           'geocentrictrueecliptic', 'supergalactic', 'icrs']}
valid_coordsys['DS9'] += [f'wcs{x}' for x in string.ascii_lowercase]

# Map astropy's coordinate frame names with their respective name in
# the file format.
coordsys_mapping = {'DS9': {x: x for x in valid_coordsys['DS9']},
                    'CRTF': {x: x.upper() for x in valid_coordsys['CRTF']}}

# Coordinate frame names must be uppercase following the CASA CRTF syntax
coordsys_mapping['CRTF']['geocentrictrueecliptic'] = 'ECLIPTIC'
coordsys_mapping['CRTF']['fk5'] = 'J2000'
coordsys_mapping['CRTF']['fk4'] = 'B1950'
coordsys_mapping['CRTF']['supergalactic'] = 'SUPERGAL'

coordsys_mapping['DS9']['geocentrictrueecliptic'] = 'ECLIPTIC'


class RegionConversionError(ValueError):
    """
    A generic error class for Shape to Region conversions.
    """


class _ShapeList(list):
    """
    A class to hold a list of `~regions.Shape` objects.
    """

    def to_regions(self):
        """
        Convert to a list of `~regions.Region` objects.
        """
        regions = []
        for shape in self:
            # Skip elliptical multi-annulus for now
            if shape.region_type == 'ellipse' and len(shape.coord) > 5:
                msg = f'Skipping elliptical annulus {shape}'
                warn(msg, DS9RegionParserWarning)
                continue

            # Skip DS9 circular multi-annulus for now
            if shape.region_type == 'circleannulus' and len(shape.coord) > 4:
                msg = f'Skipping circular annulus {shape}'
                warn(msg, DS9RegionParserWarning)
                continue

            if shape.region_type in ['box'] and shape.format_type == 'CRTF':
                msg = f'Skipping {shape.region_type} {shape}'
                warn(msg, CRTFRegionParserWarning)
                continue

            region = shape.to_region()
            regions.append(region)

        return regions

    def to_crtf(self, coordsys='fk5', fmt='.6f', radunit='deg'):
        """
        Convert to CRTF (CASA Region Text Format) region strings.

        Parameters
        ----------
        coordsys : str
            An Astropy coordinate system that overrides the coordinate
            system frame for all regions.

        fmt : str
            A python string format defining the output precision.
            Default is '.6f', which is accurate to 0.0036 arcseconds.

        radunit : str
            The unit of the radius.

        Returns
        -------
        region_string : str
            A CRFT region string.
        """
        crtf_strings = {
            'circle': '{0}circle[[{1:FMT}deg, {2:FMT}deg], {3:FMT}RAD]',
            'circleannulus': ('{0}annulus[[{1:FMT}deg, {2:FMT}deg], '
                              '[{3:FMT}RAD, {4:FMT}RAD]]'),
            # Make sure that width goes to minor axis and height to
            # major axis
            'ellipse': ('{0}ellipse[[{1:FMT}deg, {2:FMT}deg], [{4:FMT}RAD, '
                        '{3:FMT}RAD], {5:FMT}deg]'),
            'rectangle': ('{0}rotbox[[{1:FMT}deg, {2:FMT}deg], [{3:FMT}RAD, '
                          '{4:FMT}RAD], {5:FMT}deg]'),
            'polygon': '{0}poly[{1}]',
            'point': '{0}point[[{1:FMT}deg, {2:FMT}deg]]',
            'symbol': '{0}symbol[[{1:FMT}deg, {2:FMT}deg], {symbol}]',
            'text': '{0}text[[{1:FMT}deg, {2:FMT}deg], \'{text}\']',
            'line': ('{0}line[[{1:FMT}deg, {2:FMT}deg], [{3:FMT}deg, '
                     '{4:FMT}deg]]')}

        output = '#CRTFv0\n'

        if radunit == 'arcsec':
            # arcseconds are allowed for all but image coordinates
            if coordsys.lower() not in ('image',):
                radunitstr = '"'
            else:
                raise ValueError('Radius unit arcsec not valid for '
                                 f'coordsys {coordsys}')
        else:
            radunitstr = radunit

        for key, val in crtf_strings.items():
            crtf_strings[key] = val.replace('FMT', fmt).replace('RAD',
                                                                radunitstr)

        # CASA does not support global coordinate specification, even
        # though the documentation for the specification explicitly
        # states that it does.
        global_coord = coordsys_mapping['CRTF'][coordsys.lower()]
        output += f'global coord={global_coord}\n'

        for shape in self:
            shape.check_crtf()
            shape.meta = _to_crtf_meta(shape.meta)

            # if unspecified, include is True. Despite the
            # specification, CASA does *not* support a preceding "+". If
            # you want a region included, leave the opening character
            # blank.
            include = ''
            if shape.include in (False, '-'):
                include = '-'

            if shape.meta.get('type', 'reg') == 'ann':
                include += 'ann '

            if shape.meta.get('label', '') != '':
                shape.meta['label'] = f"'{shape.meta['label']}'"

            keylist = ('include', 'comment', 'symbol', 'coord', 'text',
                       'range', 'corr', 'type')
            meta_pairs = []
            for key, val in shape.meta.items():
                if key not in keylist:
                    meta_pairs.append(f'{key}={val}')
            meta_str = ', '.join(meta_pairs)

            # The first item should be the coordinates, since CASA
            # cannot recognize a region without an inline coordinate
            # specification. It can be, but does not need to be,
            # comma-separated at the start.
            shape_coordsys = getattr(shape, 'coordsys')
            if shape_coordsys.lower() != coordsys.lower():
                coord = coordsys_mapping['CRTF'][coordsys.lower()]
                if meta_str.strip():
                    meta_str = f'coord={coord}, ' + meta_str
                else:
                    # if there is no metadata at all (above), the
                    # trailing comma is incorrect
                    meta_str = f'coord={coord}'

            if 'comment' in shape.meta:
                meta_str += ', ' + shape.meta['comment']

            if 'range' in shape.meta:
                shape.meta['range'] = [str(str(x).replace(' ', '')) for x in
                                       shape.meta['range']]
                meta_str += f", range={shape.meta['range']}".replace("'", "")
            if 'corr' in shape.meta:
                meta_str += f", corr={shape.meta['corr']}".replace("'", "")

            coord = []
            if coordsys not in ['image', 'physical']:
                for val in shape.coord:
                    if isinstance(val, Angle):
                        coord.append(float(val.value))
                    else:
                        if radunit == '' or radunit is None:
                            coord.append(float(val.value))
                        else:
                            coord.append(float(val.to(radunit).value))
            else:
                for val in shape.coord:
                    if isinstance(val, u.Quantity):
                        coord.append(float(val.value))
                    else:
                        coord.append(float(val))

            if (shape.region_type in ['ellipse', 'rectangle']
                    and len(shape.coord) % 2 == 1):
                coord[-1] = float(shape.coord[-1].to('deg').value)

            if shape.region_type == 'polygon':
                vals = [f'[{x:{fmt}}deg, {y:{fmt}}deg]'
                        for x, y in zip(coord[::2], coord[1::2])]
                coord = ", ".join(vals)
                line = crtf_strings['polygon'].format(include, coord)

            elif shape.region_type == 'point':
                if 'symbol' in shape.meta:
                    line = crtf_strings['symbol'].format(
                        include, *coord, symbol=shape.meta['symbol'])
                else:
                    line = crtf_strings['point'].format(include, *coord)

            elif shape.region_type == 'ellipse':
                coord[2:] = [x / 2 for x in coord[2:]]
                if len(coord) % 2 == 1:
                    coord[-1] *= 2
                line = crtf_strings['ellipse'].format(include, *coord)

            elif shape.region_type == 'text':
                line = crtf_strings['text'].format(
                    include, *coord, text=shape.meta['text'])
            else:
                line = crtf_strings[shape.region_type].format(include, *coord)

            if meta_str.strip():
                output += f"{line}, {meta_str}\n"
            else:
                output += f"{line}\n"

        return output

    def to_ds9(self, coordsys='fk5', fmt='.6f', radunit='deg'):
        """
        Convert to DS9 region strings.

        Parameters
        ----------
        coordsys : str
            An Astropy coordinate system that overrides the coordinate
            system frame for all regions.

        fmt : str
            A python string format defining the output precision.
            Default is '.6f', which is accurate to 0.0036 arcseconds.

        radunit : str
            The unit of the radius.

        Returns
        -------
        region_string : str
            A DS9 region string.
        """
        valid_symbols_reverse = {y: x for x, y in valid_symbols_ds9.items()}

        ds9_strings = {
            'circle': '{0}circle({1:FMT},{2:FMT},{3:FMT}RAD)',
            'circleannulus': ('{0}annulus({1:FMT},{2:FMT},{3:FMT}RAD,'
                              '{4:FMT}RAD)'),
            'ellipse': ('{0}ellipse({1:FMT},{2:FMT},{3:FMT}RAD,{4:FMT}RAD,'
                        '{5:FMT})'),
            'rectangle': ('{0}box({1:FMT},{2:FMT},{3:FMT}RAD,{4:FMT}RAD,'
                          '{5:FMT})'),
            'polygon': '{0}polygon({1})',
            'point': '{0}point({1:FMT},{2:FMT})',
            'line': '{0}line({1:FMT},{2:FMT},{3:FMT},{4:FMT})',
            'text': '{0}text({1:FMT},{2:FMT})'}

        output = '# Region file format: DS9 astropy/regions\n'

        if radunit == 'arcsec':
            radunitstr = '"'
        else:
            radunitstr = ''

        for key, val in ds9_strings.items():
            ds9_strings[key] = val.replace('FMT', fmt).replace('RAD',
                                                               radunitstr)

        output += f'{coordsys}\n'

        for shape in self:
            shape.check_ds9()
            shape.meta = _to_ds9_meta(shape.meta)

            # if unspecified, include is True.
            include = ''
            if shape.include in (False, '-'):
                include = '-'

            if 'point' in shape.meta:
                shape.meta['point'] = \
                    valid_symbols_reverse[shape.meta['point']]

            if 'symsize' in shape.meta:
                shape.meta['point'] += f' {shape.meta.pop("symsize")}'

            keylist = ('include', 'tag', 'comment', 'font', 'text')
            meta_pairs = []
            for key, val in shape.meta.items():
                if key not in keylist:
                    meta_pairs.append(f'{key}={val}')
            meta_str = ' '.join(meta_pairs)

            if 'tag' in shape.meta:
                tags = [f'tag={tag}' for tag in shape.meta['tag']]
                meta_str += ' ' + ' '.join(tags)

            if 'font' in shape.meta:
                meta_str += f" font=\"{shape.meta['font']}\""

            if shape.meta.get('text', '') != '':
                meta_str += ' text={' + shape.meta['text'] + '}'

            if 'comment' in shape.meta:
                meta_str += f' {shape.meta["comment"]}'

            coord = []
            if coordsys not in ['image', 'physical']:
                for val in shape.coord:
                    if isinstance(val, Angle):
                        coord.append(float(val.value))
                    else:
                        if radunit == '' or radunit is None:
                            coord.append(float(val.value))
                        else:
                            coord.append(float(val.to(radunit).value))

                if (shape.region_type in ['ellipse', 'rectangle']
                        and len(shape.coord) % 2 == 1):
                    coord[-1] = float(shape.coord[-1].to('deg').value)

            else:
                for val in shape.coord:
                    if isinstance(val, u.Quantity):
                        coord.append(float(val.value))
                    else:
                        coord.append(float(val))
                if shape.region_type in ['polygon', 'line']:
                    coord = [x + 1 for x in coord]
                else:
                    coord[0] += 1
                    coord[1] += 1

            if shape.region_type == 'polygon':
                coord = ",".join([f'{x:{fmt}}' for x in coord])
                line = ds9_strings['polygon'].format(include, coord)

            elif shape.region_type == 'ellipse':
                coord[2:] = [x / 2 for x in coord[2:]]
                if len(coord) % 2 == 1:
                    coord[-1] *= 2
                line = ds9_strings['ellipse'].format(include, *coord)

            else:
                line = ds9_strings[shape.region_type].format(include, *coord)

            if meta_str.strip():
                output += f'{line} # {meta_str}\n'
            else:
                output += f'{line}\n'

        return output

    def to_fits(self):
        """
        Convert to a `~astropy.table.Table` object.
        """
        max_length_coord = 1
        coord_x = []
        coord_y = []
        shapes = []
        radius = []
        rotangle_deg = []
        components = []

        reg_reverse_mapping = {value: key for key, value in
                               reg_mapping['FITS_REGION'].items()}
        reg_reverse_mapping['rectangle'] = 'ROTBOX'
        reg_reverse_mapping['circleannulus'] = 'ANNULUS'
        reg_reverse_mapping['ellipseannulus'] = 'ELLIPTANNULUS'

        for num, shape in enumerate(self):
            shapes.append(reg_reverse_mapping[shape.region_type])
            if shape.region_type == 'polygon':
                max_length_coord = max(len(shape.coord) / 2, max_length_coord)
                coord = [x.value for x in shape.coord]
                coord_x.append(coord[::2])
                coord_y.append(coord[1::2])
                radius.append(0)
                rotangle_deg.append(0)
            else:
                coord_x.append(shape.coord[0].value)
                coord_y.append(shape.coord[1].value)
                if shape.region_type in ['circle', 'circleannulus', 'point']:
                    radius.append([float(val) for val in shape.coord[2:]])
                    rotangle_deg.append(0)
                else:
                    radius.append([float(x) for x in shape.coord[2:-1]])
                    rotangle_deg.append(shape.coord[-1].to('deg').value)

            tag = shape.meta.get('tag', '')
            if tag.isdigit():
                components.append(int(tag))
            else:
                components.append(num + 1)

        # pad every value with zeros at the end to make sure that all
        # values in the column have same length
        for i in range(len(self)):
            if np.isscalar(coord_x[i]):
                coord_x[i] = np.array([coord_x[i]])
            if np.isscalar(coord_y[i]):
                coord_y[i] = np.array([coord_y[i]])
            if np.isscalar(radius[i]):
                radius[i] = np.array([radius[i]])

            coord_x[i] = np.pad(coord_x[i],
                                (0, int(max_length_coord - len(coord_x[i]))),
                                'constant', constant_values=(0, 0))
            coord_y[i] = np.pad(coord_y[i],
                                (0, int(max_length_coord - len(coord_y[i]))),
                                'constant', constant_values=(0, 0))
            radius[i] = np.pad(radius[i], (0, 4 - len(radius[i])), 'constant',
                               constant_values=(0, 0))

        table = Table([coord_x, coord_y, shapes, radius, rotangle_deg,
                       components],
                      names=('X', 'Y', 'SHAPE', 'R', 'ROTANG', 'COMPONENT'))
        table['X'].unit = 'pix'
        table['Y'].unit = 'pix'
        table['R'].unit = 'pix'
        table['ROTANG'].unit = 'deg'

        return table


class _Shape:
    """
    Helper class to represent a DS9/CRTF Region.

    This serves as intermediate step in the parsing process.

    Parameters
    ----------
    coordsys : str
        An Astropy Coordinate system frame used in the region.

    region_type : str
        The type of the region (as defined in this package).

    coord : list of `~astropy.coordinates.Angle` or `~astropy.units.Quantity`
        The region coordinates.

    meta : dict
        The meta attributes.

    composite : bool
        Flag indicting wheter the region is a Composite region.

    include : bool
        Flag indicating where to include or exclude the region.
    """

    shape_to_sky_region = dict(circle=CircleSkyRegion,
                               ellipse=EllipseSkyRegion,
                               rectangle=RectangleSkyRegion,
                               polygon=PolygonSkyRegion,
                               circleannulus=CircleAnnulusSkyRegion,
                               ellipseannulus=EllipseAnnulusSkyRegion,
                               rectangleannulus=RectangleAnnulusSkyRegion,
                               line=LineSkyRegion,
                               point=PointSkyRegion,
                               text=TextSkyRegion)

    shape_to_pixel_region = dict(circle=CirclePixelRegion,
                                 ellipse=EllipsePixelRegion,
                                 rectangle=RectanglePixelRegion,
                                 polygon=PolygonPixelRegion,
                                 circleannulus=CircleAnnulusPixelRegion,
                                 ellipseannulus=EllipseAnnulusPixelRegion,
                                 rectangleannulus=RectangleAnnulusPixelRegion,
                                 line=LinePixelRegion,
                                 point=PointPixelRegion,
                                 text=TextPixelRegion)

    error = RegionConversionError

    def __init__(self, coordsys, region_type, coord, meta, composite, include):
        self._coordsys = coordsys
        self._region_type = region_type
        self.coord = coord
        self.meta = meta
        self.composite = composite
        self.include = include

    @property
    def coordsys(self):
        return self._coordsys

    @coordsys.setter
    def coordsys(self, value):
        self._coordsys = value.lower()
        self._validate()

    @property
    def region_type(self):
        return self._region_type

    @region_type.setter
    def region_type(self, value):
        self._region_type = value.lower()
        self._validate()

    def __str__(self):
        ss = self.__class__.__name__
        ss += f'\nType: {self.meta.get("type", "reg")}'
        ss += f'\nCoord sys: {self.coordsys}'
        ss += f'\nRegion type: {self.region_type}'
        if self.region_type == 'symbol':
            ss += f"\nSymbol: {self.meta['symbol']}"
        if self.region_type == 'text':
            ss += f"\nText: {self.meta['text']}"
        ss += f'\nMeta: {self.meta}'
        ss += f'\nComposite: {self.composite}'
        ss += f'\nInclude: {self.include}'
        ss += '\n'
        return ss

    def convert_coords(self):
        """
        Process a list of coordinates.

        This mainly searches for a tuple of coordinates in the
        coordinate list and creates a SkyCoord or PixCoord object from
        them if appropriate for a given region type. This involves again
        some coordinate transformation, so this step could be moved to
        the parsing process.
        """
        if self.coordsys in ['image', 'physical']:
            coords = self._convert_pix_coords()
        else:
            coords = self._convert_sky_coords()

        if self.region_type == 'line':
            coords = [coords[0][0], coords[0][1]]

        if self.region_type == 'text':
            coords.append(self.meta['text'])

        return coords

    def _convert_sky_coords(self):
        """
        Convert to sky coordinates.
        """
        parsed_angles = []
        for x, y in zip(self.coord[:-1:2], self.coord[1::2]):
            if isinstance(x, Angle) and isinstance(y, Angle):
                parsed_angles.append((x, y))

        frame = frame_transform_graph.lookup_name(self.coordsys)

        if len(parsed_angles) == 0:
            raise ValueError('error parsing region')

        lon, lat = zip(*parsed_angles)
        if (hasattr(lon, '__len__') and hasattr(lat, '__len__')
                and len(lon) == 1 and len(lat) == 1):
            # force entries to be scalar if they are length-1
            lon, lat = u.Quantity(lon[0]), u.Quantity(lat[0])
        else:
            # otherwise, they are vector quantities
            lon, lat = u.Quantity(lon), u.Quantity(lat)
        sphcoords = UnitSphericalRepresentation(lon, lat)
        coords = [SkyCoord(frame(sphcoords))]

        if self.region_type != 'polygon':
            coords += self.coord[len(coords * 2):]

        return coords

    def _convert_pix_coords(self):
        """
        Convert to pixel coordinates.
        """
        if self.region_type in ['polygon', 'line']:
            # have to special-case polygon in the phys coord case
            # b/c can't typecheck when iterating as in sky coord case
            coords = [PixCoord(self.coord[0::2], self.coord[1::2])]
        else:
            temp = [_.value for _ in self.coord]
            coord = PixCoord(temp[0], temp[1])
            coords = [coord] + temp[2:]

        # The angle remains as a quantity object. Modulus check makes
        # sure that it works for ellipse/rectangle annulus.
        if (self.region_type in ['ellipse', 'rectangle']
                and len(coords) % 2 == 0):
            coords[-1] = self.coord[-1]

        return coords

    def to_region(self):
        """
        Convert to a `~regions.Region` object.
        """
        coords = self.convert_coords()
        viz_keywords = ['color', 'dash', 'dashlist', 'width', 'font',
                        'symsize', 'symbol', 'symsize', 'fontsize',
                        'fontstyle', 'usetex', 'labelpos', 'labeloff',
                        'linewidth', 'linestyle', 'point', 'textangle',
                        'fontweight', 'symthick']

        if isinstance(coords[0], SkyCoord):
            reg = self.shape_to_sky_region[self.region_type](*coords)
        elif isinstance(coords[0], PixCoord):
            reg = self.shape_to_pixel_region[self.region_type](*coords)
        else:
            self._raise_error("No central coordinate")

        reg.visual = RegionVisual()
        reg.meta = RegionMeta()

        # both 'text' and 'label' should be set to the same value, where
        # we default to the 'text' value since that is the one used by
        # ds9 regions
        label = self.meta.get('text', self.meta.get('label', ""))
        if label != '':
            reg.meta['label'] = label
        for key in self.meta:
            if key in viz_keywords:
                reg.visual[key] = self.meta[key]
            else:
                reg.meta[key] = self.meta[key]
        reg.meta['include'] = self.include

        return reg

    def _raise_error(self, msg):
        raise self.error(msg)

    def check_crtf(self):
        """
        Check for CRTF compatibility.
        """
        if self.region_type not in regions_attributes:
            raise ValueError(f'"{self.region_type}" is not a valid region '
                             'type')

        if self.coordsys not in valid_coordsys['CRTF']:
            raise ValueError(f'"{self.coordsys}" is not a valid coordinate '
                             'reference frame')

    def check_ds9(self):
        """
        Check for DS9 compatibility.
        """
        if self.region_type not in regions_attributes:
            raise ValueError(f'"{self.region_type}" is not a valid region '
                             'type')

        if self.coordsys not in valid_coordsys['DS9']:
            raise ValueError(f'"{self.coordsys}" is not a valid coordinate '
                             'reference frame')

    def _validate(self):
        """
        Check whether all the attributes of this object is valid.
        """
        if self.region_type not in regions_attributes:
            raise ValueError(f'"{self.region_type}" is not a valid region '
                             'type')

        if self.coordsys not in valid_coordsys['DS9'] + valid_coordsys['CRTF']:
            raise ValueError(f'"{self.coordsys}" is not a valid coordinate '
                             'reference frame')


def _to_shape_list(region_list, coordinate_system='fk5'):
    """
    Convert a list of regions into a `~regions.ShapeList` object.

    Parameters
    ----------
    region_list : list
        A list of `regions.Region` objects.

    coordinate_system : str, optional
        The Astropy coordinate system frame in which all the coordinates
        present in the ``region_list`` will be converted. Default is
        'fk5'.

    Returns
    -------
    shape_list: `regions.ShapeList`
        A `~regions.ShapeList` object.
    """
    shape_list = _ShapeList()

    for region in region_list:
        coord = []
        if isinstance(region, SkyRegion):
            reg_type = region.__class__.__name__[:-9].lower()
        else:
            reg_type = region.__class__.__name__[:-11].lower()

        for val in regions_attributes[reg_type]:
            coord.append(getattr(region, val))

        if reg_type == 'polygon':
            coord = region.vertices

        if coordinate_system:
            coordsys = coordinate_system
        else:
            if isinstance(region, SkyRegion):
                coordsys = coord[0].name
            else:
                coordsys = 'image'

        frame = frame_transform_graph.lookup_name(coordsys)

        new_coord = []
        for val in coord:
            if isinstance(val, Angle):
                # convert Angle to Quantity; Angle values get units
                # stripped in serialization, but Quantity gets converted
                new_coord.append(u.Quantity(val))
            elif isinstance(val, (u.Quantity, numbers.Number)):
                new_coord.append(val)
            elif isinstance(val, PixCoord):
                new_coord.append(u.Quantity(val.x, u.dimensionless_unscaled))
                new_coord.append(u.Quantity(val.y, u.dimensionless_unscaled))
            else:
                new_coord.append(Angle(val.transform_to(frame).spherical.lon))
                new_coord.append(Angle(val.transform_to(frame).spherical.lat))

        meta = dict(region.meta)
        meta.update(region.visual)

        if reg_type == 'text':
            meta['text'] = meta.get('text', meta.pop('label', ''))

        include = region.meta.pop('include', True)

        shape_list.append(_Shape(coordsys, reg_type, new_coord, meta, False,
                                 include))

    return shape_list


def _to_ds9_meta(shape_meta):
    """
    Make the metadata DS9 compatible by filtering and mapping the valid
    keys.

    Parameters
    ----------
    shape_meta : dict
        The meta attribute of a `regions.Shape` object.

    Returns
    -------
    meta : dict
        A DS9 compatible meta dictionary.
    """
    # meta keys allowed in DS9.
    valid_keys = ['symbol', 'include', 'tag', 'line', 'comment',
                  'name', 'select', 'highlite', 'fixed', 'label', 'text',
                  'edit', 'move', 'rotate', 'delete', 'source', 'background']

    # visual keys allowed in DS9
    valid_keys += ['color', 'dash', 'linewidth', 'font', 'dashlist',
                   'fill', 'textangle', 'symsize']

    # mapped to actual names in DS9
    key_mappings = {'symbol': 'point', 'linewidth': 'width', 'label': 'text'}

    meta = _to_io_meta(shape_meta, valid_keys, key_mappings)

    if 'font' in meta:
        meta['font'] += (f" {shape_meta.get('fontsize', 12)} "
                         f"{shape_meta.get('fontstyle', 'normal')} "
                         f"{shape_meta.get('fontweight', 'roman')}")

    return meta


def _to_crtf_meta(shape_meta):
    """
    Make the metadata CRTF compatible by filtering and mapping the valid
    keys.

    Parameters
    ----------
    shape_meta : dict
        A meta attribute of a `regions.Region` object.

    Returns
    -------
    meta : dict
        A CRTF compatible meta dictionary.
    """
    # please refer : https://casaguides.nrao.edu/index.php/CASA_Region_Format

    # meta keys allowed in CRTF
    valid_keys = ['label', 'include', 'frame', 'range', 'veltype',
                  'restfreq', 'coord', 'type', 'text', 'corr']

    # visual keys allowed in CRTF
    valid_keys += ['color', 'width', 'font', 'symthick', 'symsize',
                   'fontsize', 'fontstyle', 'usetex', 'labelpos', 'labeloff',
                   'linewidth', 'linestyle', 'symbol']

    key_mappings = {}

    return _to_io_meta(shape_meta, valid_keys, key_mappings)


def _to_io_meta(shape_meta, valid_keys, key_mappings):
    """
    Make metadata compatible with a specific IO by filtering and mapping
    to its valid keys.

    Parameters
    ----------
    shape_meta : dict
        A meta attribute of a `regions.Region` object.

    valid_keys : list
        The valid keys of a particular file format.

    key_mappings : dict
        A dictionary mapping of the actual name of the key in the
        format.

    Returns
    -------
    meta : dict
        An IO compatible meta dictionary according to ``valid_keys`` and
        ``key_mappings``.
    """
    meta = dict()

    for key in shape_meta:
        if key in valid_keys:
            meta[key_mappings.get(key, key)] = shape_meta[key]

    return meta


@deprecated('0.5', alternative='`regions.Regions`')
class ShapeList(_ShapeList):
    """
    A class to hold a list of `~regions.Shape` objects.
    """


@deprecated('0.5', alternative='`regions.Region`')
class Shape(_Shape):
    """
    Helper class to represent a DS9/CRTF Region.

    This serves as intermediate step in the parsing process.

    Parameters
    ----------
    coordsys : str
        An Astropy Coordinate system frame used in the region.

    region_type : str
        The type of the region (as defined in this package).

    coord : list of `~astropy.coordinates.Angle` or `~astropy.units.Quantity`
        The region coordinates.

    meta : dict
        The meta attributes.

    composite : bool
        Flag indicting wheter the region is a Composite region.

    include : bool
        Flag indicating where to include or exclude the region.
    """



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines text regions in both pixel and sky coordinates.
"""

from astropy.wcs.utils import pixel_to_skycoord, skycoord_to_pixel

from .point import PointPixelRegion, PointSkyRegion
from ..core import PixCoord

__all__ = ['TextSkyRegion', 'TextPixelRegion']


class TextPixelRegion(PointPixelRegion):
    """
    A text string in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the leftmost point of the text string.
    text : str
        The text string.
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from regions import PixCoord, TextPixelRegion, RegionVisual
        import matplotlib.pyplot as plt

        x, y = 15, 10
        textangle = '30'

        fig, ax = plt.subplots(1, 1)

        center = PixCoord(x=x, y=y)
        reg = TextPixelRegion(center=center, text="Hello World!",
                              visual=RegionVisual(textangle=textangle))
        reg.plot(ax=ax)

        plt.xlim(10, 30)
        plt.ylim(2.5, 20)
        ax.set_aspect('equal')
    """

    _params = ('center', 'text')

    def __init__(self, center, text, meta=None, visual=None):
        super().__init__(center, meta, visual)
        self.text = text

    def to_sky(self, wcs):
        center = pixel_to_skycoord(self.center.x, self.center.y, wcs=wcs)
        return TextSkyRegion(center, self.text)

    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Return a matplotlib Text object for this region
        (`matplotlib.text.Text`).

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        **kwargs : dict
            Any keyword arguments accepted by `~matplotlib.text.Text`.
            These keywords will override any visual meta attributes of
            this region.

        Returns
        -------
        artist : `~matplotlib.text.Text`
            A matplotlib Text object.
        """
        from matplotlib.text import Text

        mpl_kwargs = self._define_mpl_kwargs(artist='Text')
        mpl_kwargs.update(kwargs)

        return Text(self.center.x - origin[0], self.center.y - origin[1],
                    self.text, **mpl_kwargs)


class TextSkyRegion(PointSkyRegion):
    """
    A text string in sky coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the leftmost point of the text string.
    text : str
        The text string.
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.
    """

    _params = ('center', 'text')

    def __init__(self, center, text, meta=None, visual=None):
        super().__init__(center, meta, visual)
        self.text = text

    def to_pixel(self, wcs):
        center_x, center_y = skycoord_to_pixel(self.center, wcs=wcs)
        center = PixCoord(center_x, center_y)
        return TextPixelRegion(center, self.text, self.meta, self.visual)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines annulus regions in both pixel and sky coordinates.
"""

import abc
import operator

import astropy.units as u
from astropy.wcs.utils import pixel_to_skycoord

from ..core.attributes import (ScalarPix, ScalarLength, QuantityLength,
                               ScalarSky)
from ..core.compound import CompoundPixelRegion
from ..core.core import PixelRegion, SkyRegion
from ..core.metadata import RegionMeta, RegionVisual
from ..core.pixcoord import PixCoord
from ..shapes.circle import CirclePixelRegion
from ..shapes.ellipse import EllipsePixelRegion, EllipseSkyRegion
from ..shapes.rectangle import RectanglePixelRegion, RectangleSkyRegion
from .._utils.wcs_helpers import pixel_scale_angle_at_skycoord

__all__ = ['AnnulusPixelRegion', 'AsymmetricAnnulusPixelRegion',
           'AsymmetricAnnulusSkyRegion',
           'CircleAnnulusPixelRegion', 'CircleAnnulusSkyRegion',
           'EllipseAnnulusPixelRegion', 'EllipseAnnulusSkyRegion',
           'RectangleAnnulusPixelRegion', 'RectangleAnnulusSkyRegion']


class AnnulusPixelRegion(PixelRegion, abc.ABC):
    """
    A base class for annulus pixel regions.
    """

    @property
    def _compound_region(self):
        return CompoundPixelRegion(self._inner_region, self._outer_region,
                                   operator.xor, self.meta, self.visual)

    @property
    def area(self):
        return self._outer_region.area - self._inner_region.area

    @property
    def bounding_box(self):
        return self._outer_region.bounding_box

    def contains(self, pixcoord):
        return self._compound_region.contains(pixcoord)

    def as_artist(self, origin=(0, 0), **kwargs):
        return self._compound_region.as_artist(origin, **kwargs)

    def to_mask(self, mode="center", subpixels=5):
        return self._compound_region.to_mask(mode, subpixels)

    def rotate(self, center, angle):
        """
        Rotate the region.

        Postive ``angle`` corresponds to counter-clockwise rotation.

        Parameters
        ----------
        center : `~regions.PixCoord`
            The rotation center point.
        angle : `~astropy.coordinates.Angle`
            The rotation angle.

        Returns
        -------
        region : `~regions.AnnulusPixelRegion`
            The rotated region (which is an independent copy).
        """
        changes = {}
        changes['center'] = self.center.rotate(center, angle)
        if hasattr(self, 'angle'):
            changes['angle'] = self.angle + angle
        return self.copy(**changes)


class CircleAnnulusPixelRegion(AnnulusPixelRegion):
    """
    A circular annulus in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the annulus.
    inner_radius : float
        The inner radius of the annulus.
    outer_radius : float
        The outer radius of the annulus.
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from regions import PixCoord, CircleAnnulusPixelRegion
        import matplotlib.pyplot as plt

        x, y = 6, 6
        inner_radius = 5.5
        outer_radius = 8
        fig, ax = plt.subplots(1, 1)

        center = PixCoord(x=x, y=y)
        reg = CircleAnnulusPixelRegion(center=center,
                                       inner_radius=inner_radius,
                                       outer_radius=outer_radius)
        patch = reg.as_artist(facecolor='none', edgecolor='red', lw=2)
        ax.add_patch(patch)

        plt.xlim(-5, 20)
        plt.ylim(-5, 20)
        ax.set_aspect('equal')
    """

    _component_class = CirclePixelRegion
    _params = ('center', 'inner_radius', 'outer_radius')
    center = ScalarPix('center')
    inner_radius = ScalarLength('inner_radius')
    outer_radius = ScalarLength('outer_radius')

    def __init__(self, center, inner_radius, outer_radius, meta=None,
                 visual=None):
        self.center = center
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    @property
    def _inner_region(self):
        return self._component_class(self.center, self.inner_radius,
                                     self.meta, self.visual)

    @property
    def _outer_region(self):
        return self._component_class(self.center, self.outer_radius,
                                     self.meta, self.visual)

    def to_sky(self, wcs):
        center = pixel_to_skycoord(self.center.x, self.center.y, wcs)
        _, pixscale, _ = pixel_scale_angle_at_skycoord(center, wcs)
        inner_radius = self.inner_radius * u.pix * pixscale
        outer_radius = self.outer_radius * u.pix * pixscale
        return CircleAnnulusSkyRegion(center, inner_radius, outer_radius,
                                      self.meta, self.visual)


class CircleAnnulusSkyRegion(SkyRegion):
    """
    A circular annulus in sky coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the center of the annulus.
    inner_radius : `~astropy.units.Quantity`
        The inner radius of the annulus in angular units.
    outer_radius : `~astropy.units.Quantity`
        The outer radius of the annulus in angular units.
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.
    """

    _params = ('center', 'inner_radius', 'outer_radius')
    center = ScalarSky('center')
    inner_radius = QuantityLength('inner_radius')
    outer_radius = QuantityLength('outer_radius')

    def __init__(self, center, inner_radius, outer_radius, meta=None,
                 visual=None):
        self.center = center
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    def to_pixel(self, wcs):
        center, pixscale, _ = pixel_scale_angle_at_skycoord(self.center, wcs)
        inner_radius = (self.inner_radius / pixscale).to(u.pix).value
        outer_radius = (self.outer_radius / pixscale).to(u.pix).value
        return CircleAnnulusPixelRegion(center, inner_radius, outer_radius,
                                        self.meta, self.visual)


class AsymmetricAnnulusPixelRegion(AnnulusPixelRegion):
    """
    Helper class for asymmetric annuli sky regions.

    Used for ellipse and rectangle pixel annulus regions.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the annulus.
    center : `~regions.PixCoord`
        The position of the center of the annulus.
    inner_width : float
        The inner width of the annulus (before rotation) in pixels.
    inner_height : float
        The inner height of the annulus (before rotation) in pixels.
    outer_width : float
        The outer width of the annulus (before rotation) in pixels.
    outer_height : float
        The outer height of the annulus (before rotation) in pixels.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the annulus, measured anti-clockwise. If
        set to zero (the default), the width axis is lined up with the x
        axis.
    """

    _params = ('center', 'inner_width', 'inner_height', 'outer_width',
               'outer_height', 'angle')
    center = ScalarPix('center')
    inner_width = ScalarLength('inner_width')
    outer_width = ScalarLength('outer_width')
    inner_height = ScalarLength('inner_height')
    outer_height = ScalarLength('outer_height')
    angle = QuantityLength('angle')

    def __init__(self, center, inner_width, outer_width, inner_height,
                 outer_height, angle=0 * u.deg, meta=None, visual=None):
        self.center = center
        self.inner_width = inner_width
        self.outer_width = outer_width
        self.inner_height = inner_height
        self.outer_height = outer_height
        self.angle = angle
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    @property
    def _inner_region(self):
        return self._component_class(self.center, self.inner_width,
                                     self.inner_height, self.angle,
                                     self.meta, self.visual)

    @property
    def _outer_region(self):
        return self._component_class(self.center, self.outer_width,
                                     self.outer_height, self.angle,
                                     self.meta, self.visual)

    def to_sky_args(self, wcs):
        center = pixel_to_skycoord(self.center.x, self.center.y, wcs)
        _, pixscale, north_angle = pixel_scale_angle_at_skycoord(center, wcs)
        inner_width = (self.inner_width * u.pix * pixscale).to(u.arcsec)
        inner_height = (self.inner_height * u.pix * pixscale).to(u.arcsec)
        outer_width = (self.outer_width * u.pix * pixscale).to(u.arcsec)
        outer_height = (self.outer_height * u.pix * pixscale).to(u.arcsec)
        angle = self.angle - (north_angle - 90 * u.deg)

        return (center, inner_width, inner_height, outer_width, outer_height,
                angle)


class AsymmetricAnnulusSkyRegion(SkyRegion):
    """
    Helper class for asymmetric annuli sky regions.

    Used for ellipse and rectangle sky annulus regions.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the center of the annulus.
    inner_width : `~astropy.units.Quantity`
        The inner width of the annulus (before rotation) as an angle.
    inner_height : `~astropy.units.Quantity`
        The inner height of the annulus (before rotation) as an angle.
    outer_width : `~astropy.units.Quantity`
        The outer width of the annulus (before rotation) as an angle.
    outer_height : `~astropy.units.Quantity`
        The outer height of the annulus (before rotation) as an angle.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the annulus, measured anti-clockwise. If
        set to zero (the default), the width axis is lined up with the
        longitude axis of the celestial coordinates
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.
    """

    _params = ('center', 'inner_width', 'inner_height', 'outer_width',
               'outer_height', 'angle')

    center = ScalarSky('center')
    inner_width = QuantityLength('inner_width')
    outer_width = QuantityLength('outer_width')
    inner_height = QuantityLength('inner_height')
    outer_height = QuantityLength('outer_height')
    angle = QuantityLength('angle')

    def __init__(self, center, inner_width, outer_width, inner_height,
                 outer_height, angle=0 * u.deg, meta=None, visual=None):
        self.center = center
        self.inner_width = inner_width
        self.outer_width = outer_width
        self.inner_height = inner_height
        self.outer_height = outer_height
        self.angle = angle
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    def to_pixel_args(self, wcs):
        center, pixscale, north_angle = pixel_scale_angle_at_skycoord(
            self.center, wcs)
        center = PixCoord(center.x, center.y)
        inner_width = (self.inner_width / pixscale).to(u.pix).value
        inner_height = (self.inner_height / pixscale).to(u.pix).value
        outer_width = (self.outer_width / pixscale).to(u.pix).value
        outer_height = (self.outer_height / pixscale).to(u.pix).value
        angle = self.angle + (north_angle - 90 * u.deg)

        return (center, inner_width, inner_height, outer_width, outer_height,
                angle)


class EllipseAnnulusPixelRegion(AsymmetricAnnulusPixelRegion):
    """
    A elliptical annulus in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the elliptical annulus.
    inner_width : float
        The inner width of the elliptical annulus (before rotation) in
        pixels.
    inner_height : float
        The inner height of the elliptical annulus (before rotation) in
        pixels.
    outer_width : float
        The outer width of the elliptical annulus (before rotation) in
        pixels.
    outer_height : float
        The outer height of the elliptical annulus (before rotation) in
        pixels.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the elliptical annulus, measured
        anti-clockwise. If set to zero (the default), the width axis is
        lined up with the x axis.
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from astropy.coordinates import Angle
        from regions import PixCoord, EllipseAnnulusPixelRegion
        import matplotlib.pyplot as plt

        x, y = 6, 6
        inner_width = 5.5
        inner_height = 3.5
        outer_width = 8.5
        outer_height = 6.5
        angle = Angle("45deg")

        fig, ax = plt.subplots(1, 1)

        center = PixCoord(x=x, y=y)
        reg = EllipseAnnulusPixelRegion(center=center, inner_width=inner_width,
                                        inner_height=inner_height,
                                        outer_width=outer_width,
                                        outer_height=outer_height, angle=angle)
        patch = reg.as_artist(facecolor='none', edgecolor='red', lw=2)
        ax.add_patch(patch)

        plt.xlim(-5, 20)
        plt.ylim(-5, 20)
        ax.set_aspect('equal')
    """

    _component_class = EllipsePixelRegion

    def to_sky(self, wcs):
        return EllipseAnnulusSkyRegion(
            *self.to_sky_args(wcs), meta=self.meta, visual=self.visual)


class EllipseAnnulusSkyRegion(AsymmetricAnnulusSkyRegion):
    """
    A elliptical annulus in `~astropy.coordinates.SkyCoord` coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the center of the elliptical annulus.
    inner_width : `~astropy.units.Quantity`
        The inner width of the elliptical annulus (before rotation) as
        an angle.
    inner_height : `~astropy.units.Quantity`
        The inner height of the elliptical annulus (before rotation) as
        an angle.
    outer_width : `~astropy.units.Quantity`
        The outer width of the elliptical annulus (before rotation) as
        an angle.
    outer_height : `~astropy.units.Quantity`
        The outer height of the elliptical annulus (before rotation) as
        an angle.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the elliptical annulus, measured
        anti-clockwise. If set to zero (the default), the width axis is
        lined up with the longitude axis of the celestial coordinates
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.
    """

    _component_class = EllipseSkyRegion

    def to_pixel(self, wcs):
        return EllipseAnnulusPixelRegion(
            *self.to_pixel_args(wcs), meta=self.meta, visual=self.visual)


class RectangleAnnulusPixelRegion(AsymmetricAnnulusPixelRegion):
    """
    A rectangular annulus in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the rectangular annulus.
    inner_width : float
        The inner width of the rectangular annulus (before rotation) in
        pixels.
    inner_height : float
        The inner height of the rectangular annulus (before rotation) in
        pixels.
    outer_width : float
        The outer width of the rectangular annulus (before rotation) in
        pixels.
    outer_height : float
        The outer height of the rectangular annulus (before rotation) in
        pixels.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the rectangular annulus, measured
        anti-clockwise. If set to zero (the default), the width axis is
        lined up with the x axis.
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from astropy.coordinates import Angle
        from regions import PixCoord, RectangleAnnulusPixelRegion
        import matplotlib.pyplot as plt

        x, y = 6, 6
        inner_width = 5.5
        inner_height = 3.5
        outer_width = 8.5
        outer_height = 6.5
        angle = Angle("45deg")

        fig, ax = plt.subplots(1, 1)

        center = PixCoord(x=x, y=y)
        reg = RectangleAnnulusPixelRegion(center=center,
                                          inner_width=inner_width,
                                          inner_height=inner_height,
                                          outer_width=outer_width,
                                          outer_height=outer_height,
                                          angle=angle)
        patch = reg.as_artist(facecolor='none', edgecolor='red', lw=2)
        ax.add_patch(patch)

        plt.xlim(-5, 20)
        plt.ylim(-5, 20)
        ax.set_aspect('equal')
    """

    _component_class = RectanglePixelRegion

    def to_sky(self, wcs):
        return RectangleAnnulusSkyRegion(
            *self.to_sky_args(wcs), meta=self.meta, visual=self.visual)


class RectangleAnnulusSkyRegion(AsymmetricAnnulusSkyRegion):
    """
    A rectangular annulus in `~astropy.coordinates.SkyCoord`
    coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the center of the rectangular annulus.
    inner_width : `~astropy.units.Quantity`
        The inner width of the rectangular annulus (before rotation) as
        an angle.
    inner_height : `~astropy.units.Quantity`
        The inner height of the rectangular annulus (before rotation) as
        an angle.
    outer_width : `~astropy.units.Quantity`
        The outer width of the rectangular annulus (before rotation) as
        an angle.
    outer_height : `~astropy.units.Quantity`
        The outer height of the rectangular annulus (before rotation) as
        an angle.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the rectangular annulus, measured
        anti-clockwise. If set to zero (the default), the width axis is
        lined up with the longitude axis of the celestial coordinates
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.
    """

    _component_class = RectangleSkyRegion

    def to_pixel(self, wcs):
        return RectangleAnnulusPixelRegion(
            *self.to_pixel_args(wcs), meta=self.meta, visual=self.visual)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the ds9 subpackage.
"""

import os

from numpy.testing import assert_allclose
import pytest

from astropy.coordinates import Angle, SkyCoord
from astropy.tests.helper import catch_warnings, assert_quantity_allclose
import astropy.units as u
from astropy.utils.data import get_pkg_data_filename, get_pkg_data_filenames
from astropy.utils.exceptions import AstropyUserWarning

from ....core import Regions
from ....shapes.circle import CircleSkyRegion
from ..core import DS9RegionParserWarning
from ..read import _DS9Parser


def test_read():
    # Check that all test files including reference files are readable
    files = get_pkg_data_filenames('data')
    # DS9RegionParserWarning from non-supported panda, [b/e]panda regions
    with catch_warnings(DS9RegionParserWarning):
        for file in files:
            with open(file) as fh:
                _DS9Parser(fh.read(), errors='warn')


# TODO: ds9.physical.windows.reg contains different values -> Why?
@pytest.mark.parametrize('filename',
                         ['data/ds9.fk5.reg',
                          'data/ds9.fk5.hms.reg',
                          'data/ds9.fk5.hms.strip.reg',
                          'data/ds9.fk5.strip.reg',
                          'data/ds9.galactic.reg',
                          'data/ds9.galactic.hms.reg',
                          'data/ds9.galactic.hms.strip.reg',
                          'data/ds9.galactic.strip.reg',
                          'data/ds9.physical.reg',
                          'data/ds9.physical.strip.reg'])
def test_file(filename):
    filename = get_pkg_data_filename(filename)

    # DS9RegionParserWarnings from skipped multi-annulus regions
    with catch_warnings(DS9RegionParserWarning):
        regs = Regions.read(filename, errors='warn', format='ds9')

    coordsys = os.path.basename(filename).split(".")[1]

    radunit = None if coordsys == 'physical' else 'arcsec'
    actual = regs.serialize(format='ds9', coordsys=str(coordsys), fmt='.2f',
                            radunit=radunit).strip()

    strip = '_strip' if 'strip' in filename else ''
    filepath = os.path.join('data', f'{coordsys}{strip}_reference.reg')
    reffile = get_pkg_data_filename(filepath)
    with open(reffile, 'r') as fh:
        desired = fh.read().strip()

    # since metadata is not required to preserve order, we have to do a more
    # complex comparison
    desired_lines = [set(line.split(' ')) for line in desired.split('\n')]
    actual_lines = [set(line.split(' ')) for line in actual.split('\n')]
    for split_line in actual_lines:
        assert split_line in desired_lines

    for split_line in desired_lines:
        assert split_line in actual_lines


def test_ds9_serialize():
    """
    Simple test case for serialize.
    """
    center = SkyCoord(42, 43, unit='deg')
    radius = Angle(3, 'deg')
    region = CircleSkyRegion(center, radius)
    expected = ('# Region file format: DS9 astropy/regions\nfk5\n'
                'circle(42.0000,43.0000,3.0000)\n')
    actual = region.serialize(format='ds9', fmt='.4f')
    assert actual == expected


def test_ds9_string_to_objects():
    """
    Simple test case for ds9_string_to_objects.
    """
    ds9_str = ('# Region file format: DS9 astropy/regions\nfk5\n'
               'circle(42.0000,43.0000,3.0000)\n')
    regions = Regions.parse(ds9_str, format='ds9')
    reg = regions[0]

    assert_allclose(reg.center.ra.deg, 42)
    assert_allclose(reg.center.dec.deg, 43)
    assert_allclose(reg.radius.value, 3)


def test_ds9_string_to_objects_whitespace():
    """
    Simple test case for ds9_string_to_objects.
    """
    ds9_str = ('# Region file format: DS9 astropy/regions\nfk5\n'
               ' -circle(42.0000,43.0000,3.0000)\n')
    region = Regions.parse(ds9_str, format='ds9')[0]
    assert not region.meta['include']


def test_ds9_io(tmpdir):
    """
    Simple test case for write and read.
    """
    center = SkyCoord(42, 43, unit='deg', frame='fk5')
    radius = Angle(3, 'deg')
    reg = CircleSkyRegion(center, radius)
    reg.meta['name'] = 'MyName'

    filename = os.path.join(str(tmpdir), 'ds9.reg')
    reg.write(filename, coordsys='fk5', format='ds9')
    reg2 = Regions.read(filename, format='ds9')[0]

    assert_allclose(reg2.center.ra.deg, 42)
    assert_allclose(reg2.center.dec.deg, 43)
    assert_allclose(reg2.radius.value, 3)
    assert 'name' in reg2.meta
    assert reg2.meta['name'] == 'MyName'


def test_missing_region_warns():
    ds9_str = ('# Region file format: DS9 astropy/regions\nfk5\n'
               'circle(42.0000,43.0000,3.0000)\nnotaregiontype(blah)')

    # this will warn on both the commented first line and the
    # not_a_region line
    with catch_warnings(AstropyUserWarning) as ASWarn:
        regions = Regions.parse(ds9_str, format='ds9', errors='warn')

    assert len(regions) == 1
    assert len(ASWarn) == 1
    estr = ('Region type "notaregiontype" was found, but it is not one '
            'of the supported region types.')
    assert estr in str(ASWarn[0].message)


def test_global_parser():
    """
    Check that the global_parser does what's expected
    """
    global_test_str = ('global color=green dashlist=8 3 width=1 '
                       'font="helvetica 10 normal roman" select=1 '
                       'highlite=1 dash=0 fixed=0 edit=1 move=1 '
                       'delete=1 include=1 source=1')
    global_parser = _DS9Parser(global_test_str)

    exp = {'dash': '0', 'source': '1', 'move': '1',
           'font': 'helvetica 10 normal roman', 'dashlist': '8 3',
           'include': True, 'highlite': '1', 'color': 'green', 'select': '1',
           'fixed': '0', 'width': '1', 'edit': '1', 'delete': '1'}
    assert dict(global_parser.global_meta) == exp


def test_ds9_color():
    """
    Color parsing test
    """
    ds9_str = ('# Region file format: DS9 astropy/regions\nfk5\n'
               'circle(42.0000,43.0000,3.0000) # color=green\n'
               'circle(43.0000,43.0000,3.0000) # color=orange\n')

    regions = Regions.parse(ds9_str, format='ds9')
    assert regions[0].visual['color'] == 'green'
    assert regions[1].visual['color'] == 'orange'


def test_ds9_color_override_global():
    """
    Color parsing test in the presence of a global
    """
    global_test_str = ('global color=blue dashlist=8 3 width=1 '
                       'font="helvetica 10 normal roman" select=1 '
                       'highlite=1 dash=0 fixed=0 edit=1 move=1 '
                       'delete=1 include=1 source=1\n')

    reg1str = 'circle(42.0000,43.0000,3.0000) # color=green'
    reg2str = 'circle(43.0000,43.0000,3.0000) # color=orange'
    reg3str = 'circle(43.0000,43.0000,3.0000)'

    global_str = (global_test_str + 'fk5\n'
                  + '\n'.join([reg1str, reg2str, reg3str]))
    ds9_str = f'# Region file format: DS9 astropy/regions\n{global_str}\n'
    regions = Regions.parse(ds9_str, format='ds9')
    assert regions[0].visual['color'] == 'green'
    assert regions[1].visual['color'] == 'orange'
    assert regions[2].visual['color'] == 'blue'


def test_issue134_regression():
    regstr = 'galactic; circle(+0:14:26.064,+0:00:45.206,30.400")'
    regions = Regions.parse(regstr, format='ds9')
    assert regions[0].radius.value == 30.4


def test_issue65_regression():
    regstr = 'J2000; circle 188.5557102 12.0314056 1" # color=red'
    region = Regions.parse(regstr, format='ds9')[0]
    assert region.center.ra.value == 188.5557102
    assert region.center.dec.value == 12.0314056
    assert region.radius.value == 1.0


def test_frame_uppercase():
    """
    Regression test for issue #236 (PR #237)
    """
    regstr = ('GALACTIC\ncircle(188.5557102,12.0314056,7.1245) # '
              'text={This message has both ' 'a " and : in it} textangle=30')
    region = Regions.parse(regstr, format='ds9')[0]
    assert region.center.l.value == 188.5557102
    assert region.center.b.value == 12.0314056
    assert region.radius.value == 7.1245


def test_pixel_angle():
    """
    Check whether angle in PixelRegions is a u.Quantity object.
    """
    reg_str = 'image\nbox(1.5,2,2,1,0)'
    region = Regions.parse(reg_str, format='ds9')[0]
    assert isinstance(region.angle, u.quantity.Quantity)


def test_expicit_formatting_directives():
    """
    Check whether every explicit formatting directive is supported.
    """
    reg_str = 'image\ncircle(1.5, 2, 2)'
    valid_coord = _DS9Parser(reg_str).shapes[0].coord

    reg_str_explicit = 'image\ncircle(1.5p, 2p, 2p)'
    coord = _DS9Parser(reg_str_explicit).shapes[0].coord
    assert_quantity_allclose(coord, valid_coord)

    reg_str_explicit = 'image\ncircle(1.5i, 2i, 2i)'
    coord = _DS9Parser(reg_str_explicit).shapes[0].coord
    assert_quantity_allclose(coord, valid_coord)

    reg_str = 'fk5\ncircle(1.5, 2, 2)'
    valid_coord = _DS9Parser(reg_str).shapes[0].coord

    reg_str_explicit = 'fk5\ncircle(1.5d, 2d, 2d)'
    coord = _DS9Parser(reg_str_explicit).shapes[0].coord
    assert_quantity_allclose(coord, valid_coord)

    reg_str_explicit = 'fk5\ncircle(1.5r, 2r, 2r)'
    coord = _DS9Parser(reg_str_explicit).shapes[0].coord
    for val in coord:
        assert val.unit == u.Unit('rad')

    reg_str = 'fk5\ncircle(1:20:30, 2:3:7, 2)'
    valid_coord = _DS9Parser(reg_str).shapes[0].coord

    reg_str_explicit = 'fk5\ncircle(1h20m30s, 2d3m7s, 2d)'
    coord = _DS9Parser(reg_str_explicit).shapes[0].coord
    assert_quantity_allclose(u.Quantity(coord), u.Quantity(valid_coord))


def test_text_metadata():
    """
    Regression test for issue #233: make sure that text metadata is
    parsed and stored appropriately.
    """
    reg_str = 'image\ncircle(1.5, 2, 2) # text={this_is_text}'
    regions = Regions.parse(reg_str, format='ds9')

    assert len(regions) == 1
    assert regions[0].meta['text'] == 'this_is_text'
    assert regions[0].meta['label'] == 'this_is_text'
    assert regions[0].meta['text'] == 'this_is_text'


def test_angle_serialization():
    """
    Regression test for issue #223 to ensure Angle arcsec inputs are
    correctly converted to degrees.
    """
    reg = Regions([CircleSkyRegion(SkyCoord(10, 20, unit='deg'),
                                   Angle(1, 'arcsec'))])
    regstr = reg.serialize(format='ds9')
    expected = ('# Region file format: DS9 astropy/regions\nfk5\n'
                'circle(10.000009,20.000002,0.000278)\n')
    assert regstr == expected


def test_semicolon():
    """
    Regression test for issue #238 to allow semicolons in the text
    field.
    """
    regstr = ('galactic\n'
              'circle(0.003, 0.1, 206.696") # text={S17; test} color=red\n'
              'circle(0.1, -0.5, 360.148") # text={S19} color=red')

    reg = Regions.parse(regstr, format='ds9')
    assert len(reg) == 2

    regstr = ('galactic;'
              'circle(202.4,47.2,10.9) # text={test; test; A1} color=red;'
              'circle(103.8,40.6,3.6) # text="A, \'B\', ; {C}, D"'
              ' color=orange;'
              'circle(103.8,30.6,3.6) # text="A, \'B, ; C}, D" color=orange;'
              'circle(303.8,40.6,3.6) # text=\'A, "B", ; {C}, D\' color=cyan;'
              'circle(303.8,40.6,3.11);'
              'circle(303.8,30.6,3.6) # text=\'A, "B, ; {C, D\' color=cyan;'
              'circle(503.8,40.6,3.6) # text={A, "B", ; \'C\', D}'
              ' color=yellow\n'
              'circle(503.8,30.6,3.6) # text={A, "B, ; \'C, D} color=yellow;'
              '-ellipse(104.84,68.55,7.93",3.96",158.39) # dashlist=8 3'
              ' select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1'
              ' source=1 text={Ellipse;Text} background color=#0ff width=1'
              ' font="helvetica 10 normal italic";'
              "ellipse(104.84,68.55,7.93',3.96',158.39) # dashlist=8 3"
              ' text="hello; world;";'
              'circle(1.5, 2, 2) # text={this_is_text};'
              'text(151.1,51.2) # text={text ; hi ; there;} textangle=45'
              ' color=orange font="helvetica 12 normal roman";'
              'text(151.1,51.2) # text  =  {text ; hi ; there;} textangle=45'
              ' color=orange font="helvetica 12 normal roman";'
              'circle(1.5, 2, 2) # text={text="hi;world";text;"tes\'t};'
              "circle(1.5, 2, 2) # text={text='hi;world';text;\"tes't};")

    reg = Regions.parse(regstr, format='ds9')
    assert len(reg) == 15


def test_parser_no_metadata():
    """
    Regression test for issue #259 to ensure regions without metadata
    are parsed correctly.
    """
    regstr1 = 'galactic;circle(42,43,3)'
    regstr2 = 'galactic;circle 42 43 3'
    reg1 = Regions.parse(regstr1, format='ds9')[0]
    reg2 = Regions.parse(regstr2, format='ds9')[0]

    assert isinstance(reg1, CircleSkyRegion)
    assert isinstance(reg2, CircleSkyRegion)
    assert reg1.radius.value == 3.0
    assert reg2.radius.value == 3.0
    assert reg1.radius.unit == 'deg'
    assert reg2.radius.unit == 'deg'


def test_spaces_metadata():
    """
    Regression test for spaces before or after "=" in region metadata.
    """
    regstr = ('galactic;'
              'circle(202.4,47.2,10.9) # text={a; test; test; A1} color=red;'
              'circle(202.4,47.2,10.9) # text ={a; test; test; A1} color= red;'
              'circle(202.4,47.2,10.9) # text= {a; test; test; A1} color =red;'
              'circle(202.4,47.2,10.9) # text = {a; test; test; A1}'
              ' color = red;')
    reg = Regions.parse(regstr, format='ds9')
    assert len(reg) == 4
    for i in range(1, len(reg)):
        assert reg[0].visual == reg[i].visual
        assert reg[0].meta == reg[i].meta

    regstr2 = ('galactic;'
               'text(151.1,51.2) # text={text ; hi ; there;} textangle= 45'
               ' color=orange font="helvetica 12 normal roman";'
               'text(151.1,51.2) # text  ={text ; hi ; there;} textangle =45'
               ' color=orange font="helvetica 12 normal roman";'
               'text(151.1,51.2) # text=  {text ; hi ; there;} textangle  =45'
               ' color   =   orange font="helvetica 12 normal roman";'
               'text(151.1,51.2) # text   =  {text ; hi ; there;} textangle=45'
               ' color   =orange font="helvetica 12 normal roman";')

    reg = Regions.parse(regstr2, format='ds9')
    assert len(reg) == 4
    for i in range(1, len(reg)):
        assert reg[0].visual == reg[i].visual
        assert reg[0].meta == reg[i].meta



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
from numpy.testing import assert_allclose, assert_equal
import pytest

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.data import get_pkg_data_filename
from astropy.wcs import WCS

from ...core import PixCoord
from ...tests.helpers import make_simple_wcs
from ..rectangle import RectanglePixelRegion, RectangleSkyRegion
from .test_common import BaseTestPixelRegion, BaseTestSkyRegion
from .utils import HAS_MATPLOTLIB  # noqa


@pytest.fixture(scope='session', name='wcs')
def wcs_fixture():
    filename = get_pkg_data_filename('data/example_header.fits')
    header = fits.getheader(filename)
    return WCS(header)


def test_corners():
    xc, yc = 2, 2
    angle = 30 * u.deg
    width = 2
    height = 1
    reg = RectanglePixelRegion(PixCoord(xc, yc), width=width, height=height,
                               angle=angle)

    y1 = yc + np.cos(angle) * (height / 2) + np.sin(angle) * (width / 2)
    x1 = xc + np.cos(angle) * (width / 2) - np.sin(angle) * (height / 2)

    assert (x1, y1) in reg.corners

    reg = RectanglePixelRegion(PixCoord(xc, yc), width=width, height=height,
                               angle=90 * u.deg)
    # simple case: rotate by 90
    np.testing.assert_array_equal([(2.5, 1.), (2.5, 3.), (1.5, 3.), (1.5, 1.)],
                                  reg.corners)

    reg = RectanglePixelRegion(center=PixCoord(xc, yc), width=width,
                               height=height, angle=0 * u.deg)
    # simpler case: rotate by 0
    np.testing.assert_array_equal([(1, 1.5), (3, 1.5), (3, 2.5), (1, 2.5)],
                                  reg.corners)

    poly = reg.to_polygon()
    assert len(poly.vertices) == 4


class TestRectanglePixelRegion(BaseTestPixelRegion):

    reg = RectanglePixelRegion(center=PixCoord(3, 4), width=4, height=3,
                               angle=5 * u.deg)
    sample_box = [-2, 8, -1, 9]
    inside = [(4.5, 4)]
    outside = [(5, 2.5)]
    expected_area = 12
    expected_repr = ('<RectanglePixelRegion(center=PixCoord(x=3, y=4), '
                     'width=4, height=3, angle=5.0 deg)>')
    expected_str = ('Region: RectanglePixelRegion\ncenter: PixCoord(x=3, '
                    'y=4)\nwidth: 4\nheight: 3\nangle: 5.0 deg')

    def test_copy(self):
        reg = self.reg.copy()
        assert reg.center.xy == (3, 4)
        assert reg.width == 4
        assert reg.height == 3
        assert_allclose(reg.angle.to_value("deg"), 5)
        assert reg.visual == {}
        assert reg.meta == {}

    def test_pix_sky_roundtrip(self):
        wcs = make_simple_wcs(SkyCoord(2 * u.deg, 3 * u.deg), 0.1 * u.deg, 20)
        reg_new = self.reg.to_sky(wcs).to_pixel(wcs)
        assert_allclose(reg_new.center.x, self.reg.center.x)
        assert_allclose(reg_new.center.y, self.reg.center.y)
        assert_allclose(reg_new.width, self.reg.width)
        assert_allclose(reg_new.height, self.reg.height)
        assert_quantity_allclose(reg_new.angle, self.reg.angle)

    @pytest.mark.skipif('not HAS_MATPLOTLIB')
    def test_as_artist(self):
        patch = self.reg.as_artist()
        # Note: `reg.center` is the center, `patch.xy` is the lower-left
        # corner
        assert_allclose(patch.xy, (1.138344, 2.331396), atol=1e-3)
        assert_allclose(patch.get_width(), 4)
        assert_allclose(patch.get_height(), 3)
        assert_allclose(patch.angle, 5)

    def test_rotate(self):
        reg = self.reg.rotate(PixCoord(2, 3), 90 * u.deg)
        assert_allclose(reg.center.xy, (1, 4))
        assert_allclose(reg.angle.to_value('deg'), 95)

    @pytest.mark.parametrize('sync', (False, True))
    def test_as_mpl_selector(self, sync):
        plt = pytest.importorskip('matplotlib.pyplot')

        data = np.random.random((16, 16))
        mask = np.zeros_like(data)

        ax = plt.subplot(1, 1, 1)
        ax.imshow(data)

        def update_mask(reg):
            mask[:] = reg.to_mask(
                mode='subpixels', subpixels=10).to_image(data.shape)

        # For now this will only work with unrotated rectangles. Once
        # this works with rotated rectangles, the following exception
        # check can be removed as well as the ``angle=0 * u.deg`` in the
        # call to copy() below.
        with pytest.raises(NotImplementedError,
                           match=('Cannot create matplotlib selector for '
                                  'rotated rectangle.')):
            self.reg.as_mpl_selector(ax)

        region = self.reg.copy(angle=0 * u.deg)

        selector = region.as_mpl_selector(ax, callback=update_mask, sync=sync)  # noqa

        from matplotlib.backend_bases import MouseEvent, MouseButton

        x, y = ax.transData.transform([[7.3, 4.4]])[0]
        ax.figure.canvas.callbacks.process('button_press_event',
                                           MouseEvent('button_press_event',
                                                      ax.figure.canvas, x, y,
                                                      button=MouseButton.LEFT))
        x, y = ax.transData.transform([[9.3, 5.4]])[0]
        ax.figure.canvas.callbacks.process('motion_notify_event',
                                           MouseEvent('button_press_event',
                                                      ax.figure.canvas, x, y,
                                                      button=MouseButton.LEFT))
        x, y = ax.transData.transform([[9.3, 5.4]])[0]
        ax.figure.canvas.callbacks.process('button_release_event',
                                           MouseEvent('button_press_event',
                                                      ax.figure.canvas, x, y,
                                                      button=MouseButton.LEFT))

        ax.figure.canvas.draw()

        if sync:
            assert_allclose(region.center.x, 8.3)
            assert_allclose(region.center.y, 4.9)
            assert_allclose(region.width, 2)
            assert_allclose(region.height, 1)
            assert_quantity_allclose(region.angle, 0 * u.deg)

            assert_equal(mask, region.to_mask(
                mode='subpixels', subpixels=10).to_image(data.shape))

        else:
            assert_allclose(region.center.x, 3)
            assert_allclose(region.center.y, 4)
            assert_allclose(region.width, 4)
            assert_allclose(region.height, 3)
            assert_quantity_allclose(region.angle, 0 * u.deg)

            assert_equal(mask, 0)

        with pytest.raises(Exception, match=('Cannot attach more than one '
                                             'selector to a region.')):
            region.as_mpl_selector(ax)


def test_rectangular_pixel_region_bbox():
    # odd sizes
    width = 7
    height = 3
    a = RectanglePixelRegion(PixCoord(50, 50), width=width, height=height,
                             angle=0. * u.deg)
    assert a.bounding_box.shape == (height, width)

    a = RectanglePixelRegion(PixCoord(50.5, 50.5), width=width, height=height,
                             angle=0. * u.deg)
    assert a.bounding_box.shape == (height + 1, width + 1)

    a = RectanglePixelRegion(PixCoord(50, 50), width=width, height=height,
                             angle=90. * u.deg)
    assert a.bounding_box.shape == (width, height)

    # even sizes
    width = 8
    height = 4
    a = RectanglePixelRegion(PixCoord(50, 50), width=width, height=height,
                             angle=0. * u.deg)
    assert a.bounding_box.shape == (height + 1, width + 1)

    a = RectanglePixelRegion(PixCoord(50.5, 50.5), width=width, height=height,
                             angle=0. * u.deg)
    assert a.bounding_box.shape == (height, width)

    a = RectanglePixelRegion(PixCoord(50.5, 50.5), width=width, height=height,
                             angle=90. * u.deg)
    assert a.bounding_box.shape == (width, height)


class TestRectangleSkyRegion(BaseTestSkyRegion):

    reg = RectangleSkyRegion(center=SkyCoord(3, 4, unit='deg'),
                             width=4 * u.deg, height=3 * u.deg,
                             angle=5 * u.deg)

    expected_repr = ('<RectangleSkyRegion(center=<SkyCoord (ICRS): (ra, dec) '
                     'in deg\n    (3., 4.)>, width=4.0 deg, height=3.0 deg, '
                     'angle=5.0 deg)>')
    expected_str = ('Region: RectangleSkyRegion\ncenter: <SkyCoord '
                    '(ICRS): (ra, dec) in deg\n    (3., 4.)>\nwidth: '
                    '4.0 deg\nheight: 3.0 deg\nangle: 5.0 deg')

    def test_copy(self):
        reg = self.reg.copy()
        assert_allclose(reg.center.ra.deg, 3)
        assert_allclose(reg.width.to_value('deg'), 4)
        assert_allclose(reg.height.to_value('deg'), 3)
        assert_allclose(reg.angle.to_value('deg'), 5)
        assert reg.visual == {}
        assert reg.meta == {}

    def test_contains(self, wcs):
        position = SkyCoord([1, 3] * u.deg, [2, 4] * u.deg)
        # 1,2 is outside, 3,4 is the center and is inside
        assert all(self.reg.contains(position, wcs)
                   == np.array([False, True], dtype='bool'))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines a class for region masks.
"""
import warnings

import astropy.units as u
import numpy as np

__all__ = ['RegionMask']


class RegionMask:
    """
    Class for a region mask.

    Parameters
    ----------
    data : array_like
        A 2D array representing the fractional overlap of a region
        on the pixel grid. This should be the full-sized (i.e., not
        truncated) array that is the direct output of one of the
        low-level "geometry" functions.

    bbox : `regions.BoundingBox`
        The bounding box object defining the region minimal bounding
        box.

    Examples
    --------
    Usage examples are provided in the :ref:`overlap-masks` section of
    the docs.
    """

    def __init__(self, data, bbox):
        self.data = np.asanyarray(data)
        if self.data.shape != bbox.shape:
            raise ValueError('mask data and bounding box must have the same '
                             'shape')
        self.bbox = bbox
        self._mask = (self.data == 0)

    def __array__(self):
        """
        Array representation of the mask data array (e.g., for
        matplotlib).
        """
        return self.data

    @property
    def shape(self):
        """
        The shape of the mask data array.
        """
        return self.data.shape

    def get_overlap_slices(self, shape):
        """
        Get slices for the overlapping part of the region mask and a
        2D array.

        Parameters
        ----------
        shape : 2-tuple of int
            The shape of the 2D array.

        Returns
        -------
        slices_large : tuple of slices or `None`
            A tuple of slice objects for each axis of the large array,
            such that ``large_array[slices_large]`` extracts the region
            of the large array that overlaps with the small array.
            `None` is returned if there is no overlap of the bounding
            box with the given image shape.

        slices_small : tuple of slices or `None`
            A tuple of slice objects for each axis of the region mask
            array such that ``small_array[slices_small]`` extracts the
            region that is inside the large array. `None` is returned if
            there is no overlap of the bounding box with the given image
            shape.
        """
        return self.bbox.get_overlap_slices(shape)

    def to_image(self, shape):
        """
        Return an image of the mask in a 2D array of the given shape,
        taking any edge effects into account.

        Parameters
        ----------
        shape : tuple of int
            The ``(ny, nx)`` shape of the output array.

        Returns
        -------
        result : `~numpy.ndarray`
            A 2D array of the mask.
        """
        if len(shape) != 2:
            raise ValueError('input shape must have 2 elements.')

        # find the overlap of the mask on the output image shape
        slices_large, slices_small = self.get_overlap_slices(shape)

        if slices_small is None:
            return None  # no overlap

        # insert the mask into the output image
        image = np.zeros(shape)
        image[slices_large] = self.data[slices_small]
        return image

    def cutout(self, data, fill_value=0., copy=False):
        """
        Create a cutout from the input data over the mask bounding box,
        taking any edge effects into account.

        Parameters
        ----------
        data : array_like
            A 2D array on which to apply the region mask.

        fill_value : float, optional
            The value used to fill pixels where the region mask does not
            overlap with the input ``data``.  The default is 0.

        copy : bool, optional
            If `True` then the returned cutout array will always be hold
            a copy of the input ``data``.  If `False` and the mask is
            fully within the input ``data``, then the returned cutout
            array will be a view into the input ``data``.  In cases
            where the mask partially overlaps or has no overlap with the
            input ``data``, the returned cutout array will always hold a
            copy of the input ``data`` (i.e., this keyword has no
            effect).

        Returns
        -------
        result : `~numpy.ndarray` or `None`
            A 2D array cut out from the input ``data`` representing the
            same cutout region as the region mask.  If there is a
            partial overlap of the region mask with the input data,
            pixels outside of the data will be assigned to
            ``fill_value``.  `None` is returned if there is no overlap
            of the region with the input ``data``.
        """
        data = np.asanyarray(data)
        if data.ndim != 2:
            raise ValueError('data must be a 2D array.')

        # find the overlap of the mask on the output image shape
        slices_large, slices_small = self.get_overlap_slices(data.shape)

        if slices_small is None:
            return None  # no overlap

        cutout_shape = (slices_small[0].stop - slices_small[0].start,
                        slices_small[1].stop - slices_small[1].start)

        if cutout_shape == self.shape:
            cutout = data[slices_large]
            if copy:
                # NOTE: np.copy() doesn't work with Quantity for
                # astropy 3.2.3
                cutout = cutout.copy()
            return cutout

        # cutout is always a copy for partial overlap
        if ~np.isfinite(fill_value):
            dtype = float
        else:
            dtype = data.dtype
        cutout = np.zeros(self.shape, dtype=dtype)
        cutout[:] = fill_value
        cutout[slices_small] = data[slices_large]

        if isinstance(data, u.Quantity):
            cutout <<= data.unit

        return cutout

    def multiply(self, data, fill_value=0.):
        """
        Multiply the region mask with the input data, taking any edge
        effects into account.

        The result is a mask-weighted cutout from the data.

        Parameters
        ----------
        data : array_like or `~astropy.units.Quantity`
            The 2D array to multiply with the region mask.

        fill_value : float, optional
            The value is used to fill pixels where the region mask does
            not overlap with the input ``data``.  The default is 0.

        Returns
        -------
        result : `~numpy.ndarray` or `None`
            A 2D mask-weighted cutout from the input ``data``.  If there
            is a partial overlap of the region mask with the input data,
            pixels outside of the data will be assigned to
            ``fill_value`` before being multipled with the mask.  `None`
            is returned if there is no overlap of the region with the
            input ``data``.
        """
        cutout = self.cutout(data, fill_value=fill_value)
        if cutout is None:
            return None
        else:
            # ignore multiplication with non-finite data values
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', RuntimeWarning)
                weighted_cutout = cutout * self.data

            # fill values outside of the mask but within the bounding box
            weighted_cutout[self._mask] = fill_value

            return weighted_cutout

    def get_values(self, data, mask=None):
        """
        Get the mask-weighted pixel values from the data as a 1D array.

        If the `~regions.RegionMask` was created with ``mode='center'``,
        (where the mask weights are only 1 or 0), then the returned
        values will simply be pixel values extracted from the data.

        Parameters
        ----------
        data : array_like or `~astropy.units.Quantity`
            The 2D array from which to get mask-weighted values.

        mask : array_like (bool), optional
            A boolean mask with the same shape as ``data`` where a
            `True` value indicates the corresponding element of ``data``
            is not returned in the result.

        Returns
        -------
        result : `~numpy.ndarray`
            A 1D array of mask-weighted pixel values from the input
            ``data``. If there is no overlap of the region with the
            input ``data``, the result will be an empty array with shape
            (0,).
        """
        slc_large, slc_small = self.get_overlap_slices(data.shape)
        if slc_large is None:
            return np.array([])
        cutout = data[slc_large]
        region_mask = self.data[slc_small]
        pixel_mask = (region_mask > 0)  # good pixels

        if mask is not None:
            if mask.shape != data.shape:
                raise ValueError('mask and data must have the same shape')
            pixel_mask &= ~mask[slc_large]

        # ignore multiplication with non-finite data values
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            return (cutout * region_mask)[pixel_mask]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import copy
import itertools
import re
from warnings import warn

from astropy.coordinates import Angle, frame_transform_graph
import astropy.units as u
from astropy.utils.data import get_readable_fileobj
from astropy.utils.decorators import deprecated

from ...core import Regions
from ...core.registry import RegionsRegistry
from ..core import _Shape, _ShapeList, reg_mapping
from .core import (CRTFRegionParserError, CRTFRegionParserWarning,
                   valid_symbols)

__all__ = ['read_crtf', 'CRTFParser']

# All CASA files start with '#CRTF' . It may also include the version
# number like '#CRTFv0' .
regex_begin = re.compile(r'^#CRTFv?[\d]?')

# Comment format
regex_comment = re.compile(r'^#.*$')

# Identifies the global attributes Format
regex_global = re.compile(r'^global\s+(?P<parameters>.*)?')

# Coordinate Format : '[x, y]'
regex_coordinate = re.compile(r'\[([\w.+-:]*?)\s*[,]\s*([\w.+-:]*?)\]')

# Single length format, e.g., helps extract the radius of a circle
regex_length = re.compile(r'(?:\[[^=]*\])+[,]\s*([^\[]*)\]')

# Extracts each 'parameter=value' pair
regex_meta = re.compile(r'(?:(\w+)\s*=[\s\'\"]*([^,\[\]]+?)[\'\",]+)|(?:(\w+)\s*=\s*\[(.*?)\])')  # noqa

# Region format which segregates the include ('+'|'-') parameter, the
# kind of definition ('ann' for annotations or '' for regions) and region
# type.
regex_region = re.compile(r'(?P<include>[+-])?(?P<type>ann(?=\s))?\s*(?P<regiontype>[a-z]*?)\s?\[[^=]*]')  # noqa

# Line format which checks the validity of the line and segregates the
# meta attributes from the region format.
regex_line = re.compile(r'(?P<region>[+-]?(?:ann(?=\s))?\s*[a-z]+?\s?\[[^=]+\])(?:\s*,?\s*(?P<parameters>.*))?')  # noqa


@RegionsRegistry.register(Regions, 'read', 'crtf')
def _read_crtf(filename, errors='strict', cache=False):
    """
    Read a CRTF region file and return a list of region objects.

    Parameters
    ----------
    filename : str
        The filename of the file to access.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.CRTFRegionParserError`. 'warn' will raise a
        `~regions.CRTFRegionParserWarning`, and 'ignore' will do nothing
        (i.e., be silent).

    cache : bool or 'update', optional
        Whether to cache the contents of remote URLs. If 'update', check
        the remote URL for a new version but store the result in the
        cache.

    Returns
    -------
    regions : list
        A list of `~regions.Region` objects.
    """
    with get_readable_fileobj(filename) as fh:
        if regex_begin.search(fh.readline()):
            region_string = fh.read()
            return _parse_crtf(region_string, errors=errors)
        else:
            raise CRTFRegionParserError('Every CRTF Region must start with '
                                        '"#CRTF"')


@RegionsRegistry.register(Regions, 'parse', 'crtf')
def _parse_crtf(region_string, errors='strict'):
    parser = _CRTFParser(region_string, errors=errors)
    return Regions(parser.shapes.to_regions())


class _CRTFParser:
    """
    Parse a CRTF string.

    This class transforms a CRTF string to a
    `~regions.io.core.ShapeList`. The result is stored in the ``shapes``
    attribute.

    Parameters
    ----------
    region_string : str
        A CRTF region string.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.CRTFRegionParserError`. 'warn' will raise a
        `~regions.CRTFRegionParserWarning`, and 'ignore' will do nothing
        (i.e., be silent).
    """

    # Each line is tested for either containing a region with meta
    # attributes or global parameters. If meta attributes or global
    # parameters are found, they are stored in the ``global_meta``
    # attribute. If a region is found, then ``_CRTFRegionParser``
    # is invoked to transform the line into a `~regions.io.core.Shape`
    # object, which is stored in the ``shapes`` attribute.

    # valid definition (region or annotation) types
    valid_definition = ('box', 'centerbox', 'rotbox', 'poly', 'circle',
                        'annulus', 'ellipse', 'line', 'vector', 'text',
                        'symbol')

    # valid parameters (attributes)
    valid_global_keys = ('coord', 'frame', 'corr', 'veltype', 'restfreq',
                         'linewidth', 'linestyle', 'symsize', 'symthick',
                         'color', 'font', 'fontsize', 'fontstyle', 'usetex',
                         'labelpos', 'labelcolor', 'labeloff', 'range')

    def __init__(self, region_string, errors='strict'):
        if errors not in ('strict', 'ignore', 'warn'):
            raise ValueError('errors must be one of "strict", "ignore", or '
                             '"warn"')

        self.region_string = region_string
        self.errors = errors
        self.global_meta = {}  # global states
        self.shapes = _ShapeList()
        self.run()

    def __str__(self):
        ss = self.__class__.__name__
        ss += f'\nErrors: {self.errors}'
        ss += f'\nGlobal meta: {self.global_meta}'
        ss += f'\nShapes: {self.shapes}'
        ss += '\n'
        return ss

    def parse_line(self, line):
        """
        Parse a single line.
        """
        # Skip blanks
        if line == '':
            return

        # Skip comments
        if regex_comment.search(line):
            return

        # Special case / header: parse global parameters into metadata
        global_parameters = regex_global.search(line)
        if global_parameters:
            self.parse_global_meta(global_parameters.group('parameters'))
            return

        # Tries to check the validity of the line.
        crtf_line = regex_line.search(line)
        if crtf_line:
            # Tries to parse the line.

            # Finds info about the region.
            region = regex_region.search(crtf_line.group('region'))
            type_ = region.group('type') or 'reg'
            include = region.group('include') or '+'
            region_type = region.group('regiontype').lower()
            if region_type in self.valid_definition:
                helper = _CRTFRegionParser(
                    self.global_meta, include, type_, region_type,
                    *crtf_line.group('region', 'parameters'))

                self.shapes.append(helper.shape)
            else:
                self._raise_error('Not a valid CRTF Region type: '
                                  f'"{region_type}".')

        else:
            self._raise_error(f'Not a valid CRTF line: "{line}".')
            return

    def _raise_error(self, msg):
        if self.errors == 'warn':
            warn(msg, CRTFRegionParserWarning)
        elif self.errors == 'strict':
            raise CRTFRegionParserError(msg)

    def run(self):
        """
        Run all the steps.

        This function splits the regions into lines and calls
        ``parse_line`` for each line.
        """
        for line in self.region_string.split('\n'):
            self.parse_line(line)

    def parse_global_meta(self, global_meta_str):
        """
        Parse the line starting with global to extract all the valid
        meta key/value pair.
        """
        if global_meta_str:
            global_meta_str = regex_meta.findall(global_meta_str + ',')
        if global_meta_str:
            for par in global_meta_str:
                if par[0] != '':
                    val1 = par[0].lower()
                    val2 = par[1]
                else:
                    val1 = par[2].lower()
                    val2 = par[3]
                val1 = val1.strip()
                val2 = val2.strip()
                if val1 in self.valid_global_keys:
                    if val1 in ('range', 'corr', 'labeloff'):
                        val2 = val2.split(',')
                        val2 = [x.strip() for x in val2 if x]
                    self.global_meta[val1] = val2
                else:
                    self._raise_error(f'"{val1}" is not a valid global meta '
                                      'key')


class _CRTFRegionParser:
    """
    Parse a CRTF region string.

    This will turn a line containing a CRTF region into a
    `~regions.Shape` object.

    Parameters
    ----------
    global_meta : dict
        Global meta data of the CRTF (CASA Region Text Format) file
        which is used as default meta values for regions.

    include : {'+', '-'}
        Flag at the beginning of the line.

    type_ : {'reg', 'ann'}
        Kind of the region definition.

    region_type : str
        Region type.

    reg_str : str
        Region string to parse.

    meta_str : str
        Meta string to parse.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.CRTFRegionParserError`. 'warn' will raise a
        `~regions.CRTFRegionParserWarning`, and 'ignore' will do nothing
        (i.e., be silent).
    """

    # List of valid coordinate system
    # TODO : There are still many reference systems to support
    coordinate_systems = ['j2000', 'icrs', 'galactic', 'supergal', 'image',
                          'ecliptic']

    # Maps CASA coordinate frame to appropriate astropy coordinate frames.
    coordsys_mapping = dict(zip(frame_transform_graph.get_names(),
                                frame_transform_graph.get_names()))
    coordsys_mapping['j2000'] = 'fk5'
    coordsys_mapping['b1950'] = 'fk4'
    coordsys_mapping['supergal'] = 'supergalactic'
    coordsys_mapping['ecliptic'] = 'geocentrictrueecliptic'

    # CRTF Format specifications defining how a region is read:
    #   'c' denotes a coordinates
    #   'l' denotes a length
    #   'pl' denotes a pair of lengths
    #   's' denotes a string (generally text or a symbol)
    language_spec = {'circle': ['c', 'l'],
                     'box': ['c', 'c'],
                     'centerbox': ['c', 'pl'],
                     'rotbox': ['c', 'pl', 'l'],
                     'poly': itertools.cycle('c'),
                     'annulus': ['c', 'pl'],
                     'ellipse': ['c', 'pl', 'l'],
                     'line': ['c', 'c'],
                     'vector': ['c', 'c'],
                     'symbol': ['c', 's'],
                     'text': ['c', 's']}

    def __init__(self, global_meta, include, type_, region_type, reg_str,
                 meta_str, errors='strict'):
        self.global_meta = global_meta
        self.reg_str = reg_str
        self.meta_str = meta_str
        self.errors = errors

        self.coord = None
        self.coordsys = None
        self.coord_str = None
        self.type_ = type_
        self.region_type = region_type
        self.meta = copy.deepcopy(global_meta)
        self.shape = None
        self.include = include or '+'

        self.parse()

    def _raise_error(self, msg):
        if self.errors == 'warn':
            warn(msg, CRTFRegionParserWarning)
        elif self.errors == 'strict':
            raise CRTFRegionParserError(msg)

    def parse(self):
        """
        Parse the CRTF region string.
        """
        self.convert_meta()
        self.coordsys = self.meta.get('coord', 'image').lower()
        self.set_coordsys()
        self.convert_coordinates()
        self.make_shape()

    def set_coordsys(self):
        """
        Map to astropy's coordinate system frame name.
        """
        # TODO: needs expert attention (most reference systems are not
        # mapped)
        if self.coordsys.lower() in self.coordsys_mapping:
            self.coordsys = self.coordsys_mapping[self.coordsys.lower()]

    def convert_coordinates(self):
        """
        Convert coordinate string to `~astropy.coordinates.Angle` or
        `~astropy.units.quantity.Quantity` objects.
        """
        coord_list_str = (regex_coordinate.findall(self.reg_str)
                          + regex_length.findall(self.reg_str))
        coord_list = []

        if self.region_type == 'poly':
            if len(coord_list_str) < 3:
                self._raise_error(f'Not in proper format: {self.reg_str} '
                                  'polygon should have >= 3 coordinates')
        else:
            expected_nparam = len(self.language_spec[self.region_type])
            if len(coord_list_str) != expected_nparam:
                self._raise_error(f'Not in proper format: "{self.reg_str}". '
                                  'Does not contain expected number of '
                                  'parameters for the region '
                                  f'"{self.region_type}"')

        for attr_spec, val_str in zip(self.language_spec[self.region_type],
                                      coord_list_str):
            if attr_spec == 'c':
                if len(val_str) == 2 and val_str[1] != '':
                    coord_list.append(
                        _CRTFCoordinateParser.parse_coordinate(val_str[0]))
                    coord_list.append(
                        _CRTFCoordinateParser.parse_coordinate(val_str[1]))
                else:
                    self._raise_error(f'Not in proper format: {val_str} '
                                      'should be a coordinate')

            if attr_spec == 'pl':
                if len(val_str) == 2 and val_str[1] != '':
                    coord_list.append(
                        _CRTFCoordinateParser.parse_angular_length_quantity(
                            val_str[0]))
                    coord_list.append(
                        _CRTFCoordinateParser.parse_angular_length_quantity(
                            val_str[1]))
                else:
                    self._raise_error(f'Not in proper format: {val_str} '
                                      'should be a pair of lengths')

            if attr_spec == 'l':
                if isinstance(val_str, str):
                    coord_list.append(
                        _CRTFCoordinateParser.parse_angular_length_quantity(
                            val_str))
                else:
                    self._raise_error(f'Not in proper format: {val_str} '
                                      'should be a single length')

            if attr_spec == 's':
                if self.region_type == 'symbol':
                    if val_str in valid_symbols:
                        self.meta['symbol'] = val_str
                    else:
                        self._raise_error(f'Not in proper format: "{val_str}" '
                                          'should be a symbol')
                elif self.region_type == 'text':
                    self.meta['text'] = val_str[1:-1]

        self.coord = coord_list

    def convert_meta(self):
        """
        Parse the meta_str to a dictionary and store in the ``meta``
        attribute.
        """
        if self.meta_str:
            self.meta_str = regex_meta.findall(self.meta_str + ',')
        if self.meta_str:
            for par in self.meta_str:
                if par[0] != '':
                    val1 = par[0]
                    val2 = par[1]
                else:
                    val1 = par[2]
                    val2 = par[3]
                val1 = val1.strip()
                val2 = val2.strip()
                if val1 in CRTFParser.valid_global_keys or val1 == 'label':
                    if val1 in ('range', 'corr', 'labeloff'):
                        val2 = val2.split(',')
                        val2 = [x.strip() for x in val2]
                    self.meta[val1] = val2
                else:
                    self._raise_error(f'"{val1}" is not a valid meta key')

        self.meta['include'] = self.include != '-'
        self.include = self.meta['include']

        if 'range' in self.meta:
            self.meta['range'] = [u.Quantity(x) for x in self.meta['range']]

        self.meta['type'] = self.type_

    def make_shape(self):
        """
        Make a `~regions.Shape` object.
        """
        if self.region_type == 'ellipse':
            self.coord[2:] = [x * 2 for x in self.coord[2:]]
            # Map major and minor axis to height and width respectively
            self.coord[2], self.coord[3] = self.coord[3], self.coord[2]
            if len(self.coord) % 2 == 1:  # check if angle is present
                self.coord[-1] /= 2

        if self.region_type == 'box':
            x = (self.coord[0] + self.coord[2]) / 2
            y = (self.coord[1] + self.coord[3]) / 2
            w = u.Quantity(self.coord[0] - self.coord[2])
            h = u.Quantity(self.coord[1] - self.coord[3])
            self.coord = [x, y, abs(w), abs(h)]

        self.meta.pop('coord', None)

        self.shape = _Shape(coordsys=self.coordsys,
                            region_type=reg_mapping['CRTF'][self.region_type],
                            coord=self.coord,
                            meta=self.meta,
                            composite=False,
                            include=self.include)


class _CRTFCoordinateParser:
    """
    Helper class to structure coordinate parser.
    """

    @staticmethod
    def parse_coordinate(string_rep):
        """
        Parse a single coordinate.
        """
        # Any CRTF coordinate representation (sexagesimal or degrees)
        if 'pix' in string_rep:
            return u.Quantity(string_rep[:-3], u.dimensionless_unscaled)
        if 'h' in string_rep or 'rad' in string_rep:
            return Angle(string_rep)
        if len(string_rep.split('.')) >= 3:
            string_rep = string_rep.replace('.', ':', 2)

        return Angle(string_rep, u.deg)

    @staticmethod
    def parse_angular_length_quantity(string_rep):
        """
        Parse a string into a Quantity object.

        Given a string that is a number and a unit, return a Quantity of
        that string. An error is raised if there is no unit, e.g.:

        * 50" -> 50 * u.arcsec
        * 50 -> CRTFRegionParserError : Units must be specified for 50
        """
        unit_mapping = {'deg': u.deg,
                        'rad': u.rad,
                        'arcmin': u.arcmin,
                        'arcsec': u.arcsec,
                        'pix': u.dimensionless_unscaled,
                        '"': u.arcsec,
                        "'": u.arcmin}

        regex_str = re.compile(r'([0-9+,-.]*)(.*)')
        restr = regex_str.search(string_rep)
        unit = restr.group(2)
        if unit:
            if unit in unit_mapping:
                return u.Quantity(restr.group(1), unit=unit_mapping[unit])
            return u.Quantity(restr.group(1))
        else:
            raise CRTFRegionParserError('Units must be specified for '
                                        f'{string_rep}')


@deprecated('0.5', alternative='`regions.Regions.read`')
def read_crtf(filename, errors='strict', cache=False):
    """
    Read a CRTF region file and return a list of region objects.

    Parameters
    ----------
    filename : str
        The filename of the file to access.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.CRTFRegionParserError`. 'warn' will raise a
        `~regions.CRTFRegionParserWarning`, and 'ignore' will do nothing
        (i.e., be silent).

    cache : bool or 'update', optional
        Whether to cache the contents of remote URLs. If 'update', check
        the remote URL for a new version but store the result in the
        cache.

    Returns
    -------
    regions : list
        A list of `~regions.Region` objects.
    """
    return _read_crtf(filename, errors=errors, cache=cache)


@deprecated('0.5', alternative='`regions.Regions.parse`')
class CRTFParser(_CRTFParser):
    """
    Parse a CRTF string.

    This class transforms a CRTF string to a
    `~regions.io.core.ShapeList`. The result is stored in the ``shapes``
    attribute.

    Parameters
    ----------
    region_string : str
        A CRTF region string.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.CRTFRegionParserError`. 'warn' will raise a
        `~regions.CRTFRegionParserWarning`, and 'ignore' will do nothing
        (i.e., be silent).
    """



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
The tests in this file simply check what functionality is currently
implemented and doesn't check anything about correctness.
"""
import itertools

from astropy.coordinates import SkyCoord
import astropy.units as u
from astropy.wcs import WCS
import pytest

from ...core.core import PixelRegion, Region, SkyRegion
from ...core.mask import RegionMask
from ...core.pixcoord import PixCoord
from ..annulus import (CircleAnnulusPixelRegion, CircleAnnulusSkyRegion,
                       EllipseAnnulusPixelRegion, EllipseAnnulusSkyRegion,
                       RectangleAnnulusPixelRegion, RectangleAnnulusSkyRegion)
from ..circle import CirclePixelRegion, CircleSkyRegion
from ..ellipse import EllipsePixelRegion, EllipseSkyRegion
from ..point import PointPixelRegion, PointSkyRegion
from ..polygon import PolygonPixelRegion, PolygonSkyRegion
from ..rectangle import RectanglePixelRegion, RectangleSkyRegion

PIXEL_REGIONS = [
    CirclePixelRegion(PixCoord(3, 4), radius=5),
    CircleAnnulusPixelRegion(PixCoord(3, 4), 5, 7),
    EllipsePixelRegion(PixCoord(3, 4), width=7, height=5, angle=3 * u.deg),
    EllipseAnnulusPixelRegion(PixCoord(3, 4), 7, 9, 8, 10, angle=3 * u.deg),
    PolygonPixelRegion(PixCoord([1, 4, 3], [2, 4, 4])),
    RectanglePixelRegion(PixCoord(6, 5), width=3, height=5),
    RectangleAnnulusPixelRegion(PixCoord(6, 5), 3, 6, 5, 7),
    PointPixelRegion(PixCoord(1, 2))]

SKY_REGIONS = [
    CircleSkyRegion(SkyCoord(3 * u.deg, 4 * u.deg), radius=5 * u.deg),
    CircleAnnulusSkyRegion(SkyCoord(3 * u.deg, 4 * u.deg), 5 * u.deg,
                           7 * u.deg),
    EllipseSkyRegion(SkyCoord(3 * u.deg, 4 * u.deg), width=7 * u.deg,
                     height=5 * u.deg, angle=3 * u.deg),
    EllipseAnnulusSkyRegion(SkyCoord(3 * u.deg, 4 * u.deg), 7 * u.deg,
                            9 * u.deg, 5 * u.deg, 9 * u.deg, angle=3 * u.deg),
    PolygonSkyRegion(SkyCoord([1, 4, 3] * u.deg, [2, 4, 4] * u.deg)),
    RectangleSkyRegion(SkyCoord(6 * u.deg, 5 * u.deg), width=3 * u.deg,
                       height=5 * u.deg),
    RectangleAnnulusSkyRegion(SkyCoord(6 * u.deg, 5 * u.deg), 3 * u.deg,
                              5 * u.deg, 5 * u.deg, 7 * u.deg),
    PointSkyRegion(SkyCoord(6 * u.deg, 5 * u.deg))]

MASK_MODES = ['center', 'exact', 'subpixels']
COMMON_WCS = WCS(naxis=2)
COMMON_WCS.wcs.ctype = 'RA---TAN', 'DEC--TAN'


def ids_func(arg):
    if isinstance(arg, Region):
        return arg.__class__.__name__
    else:
        return str(arg)


@pytest.mark.parametrize('region', [PIXEL_REGIONS[i] for i in (0, 2, 5)],
                         ids=ids_func)
def test_pix_in(region):
    assert PixCoord(5, 4) in region


@pytest.mark.parametrize('region', PIXEL_REGIONS, ids=ids_func)
def test_pix_area(region):
    # TODO: remove the pytest.skip once polygon area is implemented
    if isinstance(region, PolygonPixelRegion):
        pytest.skip()

    area = region.area
    assert not isinstance(area, u.Quantity)


@pytest.mark.parametrize('region', PIXEL_REGIONS, ids=ids_func)
def test_pix_to_sky(region):
    try:
        sky_region = region.to_sky(COMMON_WCS)
        assert isinstance(sky_region, SkyRegion)
    except NotImplementedError:
        pytest.xfail()


@pytest.mark.parametrize(('region', 'mode'),
                         itertools.product(PIXEL_REGIONS, MASK_MODES),
                         ids=ids_func)
def test_pix_to_mask(region, mode):
    try:
        mask = region.to_mask(mode=mode)
        assert isinstance(mask, RegionMask)
    except NotImplementedError:
        pytest.xfail()


@pytest.mark.parametrize('region', SKY_REGIONS, ids=ids_func)
def test_sky_in(region):
    region.contains(SkyCoord(1 * u.deg, 1 * u.deg, frame='icrs'), COMMON_WCS)


@pytest.mark.parametrize('region', SKY_REGIONS, ids=ids_func)
def test_sky_in_array(region):
    region.contains(SkyCoord([1, 2, 3] * u.deg, [3, 2, 1] * u.deg,
                             frame='icrs'), COMMON_WCS)


@pytest.mark.parametrize('region', SKY_REGIONS, ids=ids_func)
def test_sky_to_pix(region):
    pix_region = region.to_pixel(wcs=COMMON_WCS)
    assert isinstance(pix_region, PixelRegion)


@pytest.mark.parametrize('region', PIXEL_REGIONS, ids=ids_func)
def test_attribute_validation_pixel_regions(region):
    invalid_values = dict(center=[PixCoord([1, 2], [2, 3]), 1,
                                  SkyCoord(1 * u.deg, 2 * u.deg)],
                          radius=[u.Quantity("1deg"), [1], PixCoord(1, 2)],
                          angle=[u.Quantity([1 * u.deg, 2 * u.deg]), 2,
                                 PixCoord(1, 2)],
                          vertices=[u.Quantity("1"), 2, PixCoord(1, 2),
                                    PixCoord([[1, 2]], [[2, 3]])]
                          )
    invalid_values['width'] = invalid_values['radius']
    invalid_values['height'] = invalid_values['radius']
    invalid_values['inner_height'] = invalid_values['radius']
    invalid_values['inner_width'] = invalid_values['radius']
    invalid_values['outer_height'] = invalid_values['radius']
    invalid_values['outer_width'] = invalid_values['radius']
    invalid_values['inner_radius'] = invalid_values['radius']
    invalid_values['outer_radius'] = invalid_values['radius']
    invalid_values['start'] = invalid_values['center']
    invalid_values['end'] = invalid_values['radius']

    for attr in invalid_values:
        if hasattr(region, attr):
            for val in invalid_values.get(attr, None):
                with pytest.raises(ValueError) as excinfo:
                    setattr(region, attr, val)
                assert f'The {attr} must be' in str(excinfo.value)


@pytest.mark.parametrize('region', SKY_REGIONS, ids=ids_func)
def test_attribute_validation_sky_regions(region):
    invalid_values = dict(center=[PixCoord([1, 2], [2, 3]), 1,
                                  SkyCoord([1 * u.deg], [2 * u.deg])],
                          radius=[u.Quantity([1 * u.deg, 5 * u.deg]),
                                  [1], SkyCoord(1 * u.deg, 2 * u.deg), 1],
                          angle=[u.Quantity([1 * u.deg, 2 * u.deg]), 2,
                                 SkyCoord(1 * u.deg, 2 * u.deg)],
                          vertices=[u.Quantity("1deg"), 2,
                                    SkyCoord(1 * u.deg, 2 * u.deg),
                                    SkyCoord([[1 * u.deg, 2 * u.deg]],
                                             [[2 * u.deg, 3 * u.deg]])])

    invalid_values['width'] = invalid_values['radius']
    invalid_values['height'] = invalid_values['radius']
    invalid_values['inner_height'] = invalid_values['radius']
    invalid_values['inner_width'] = invalid_values['radius']
    invalid_values['outer_height'] = invalid_values['radius']
    invalid_values['outer_width'] = invalid_values['radius']
    invalid_values['inner_radius'] = invalid_values['radius']
    invalid_values['outer_radius'] = invalid_values['radius']
    invalid_values['start'] = invalid_values['center']
    invalid_values['end'] = invalid_values['radius']

    for attr in invalid_values:
        if hasattr(region, attr):
            for val in invalid_values.get(attr, None):
                with pytest.raises(ValueError) as excinfo:
                    setattr(region, attr, val)
                assert f'The {attr} must be' in str(excinfo.value)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the rectangular_overlap_grid module.
"""

import itertools

from numpy.testing import assert_allclose
import pytest

from .. import rectangular_overlap_grid


grid_sizes = [50, 500, 1000]
rect_sizes = [0.2, 0.4, 0.8]
angles = [0.0, 0.5, 1.0]
subsamples = [1, 5, 10]
arg_list = ['grid_size', 'rect_size', 'angle', 'subsample']


@pytest.mark.parametrize(('grid_size', 'rect_size', 'angle', 'subsample'),
                         list(itertools.product(grid_sizes, rect_sizes,
                                                angles, subsamples)))
def test_rectangular_overlap_grid(grid_size, rect_size, angle, subsample):
    """
    Test normalization of the overlap grid to make sure that a fully
    enclosed pixel has a value of 1.0.
    """

    g = rectangular_overlap_grid(-1.0, 1.0, -1.0, 1.0, grid_size, grid_size,
                                 rect_size, rect_size, angle, 0, subsample)
    assert_allclose(g.max(), 1.0)
